from __future__ import annotations

import json

import pytest
from vunnel import workspace, schema
from vunnel.providers.nvd import manager


@pytest.mark.parametrize(
    "mock_data_path",
    [
        "test-fixtures/single-entry.json",
    ],
)
def test_parser(tmpdir, helpers, mock_data_path, mocker):
    mock_data_path = helpers.local_dir(mock_data_path)

    with open(mock_data_path) as f:
        json_dict = json.load(f)

    expected_vulns = []
    for v in json_dict["vulnerabilities"]:
        cve_id = v["cve"]["id"]
        year = cve_id.split("-")[1]
        identity = f"{year}/{cve_id}"
        expected_vulns.append((identity, v))

    subject = manager.Manager(
        workspace=workspace.Workspace(tmpdir, "test", create=True),
        schema=schema.NVDSchema(),
        overrides_url="http://example.com",
    )
    subject.api.cve = mocker.Mock(return_value=[json_dict])
    actual_vulns = list(subject.get(None))

    assert expected_vulns == actual_vulns
    for vuln in actual_vulns:
        assert subject._sqlite_reader().read(vuln[0].lower()) is not None



from __future__ import annotations

import pytest

from vunnel.utils.vulnerability import build_reference_links


@pytest.mark.parametrize(
    "vulnerability_id, expected",
    [
        ("GHSA-4r7g-fj95-jwpc", ["https://github.com/advisories/GHSA-4r7g-fj95-jwpc"]),
        ("CVE-2023-1234", ["https://www.cve.org/CVERecord?id=CVE-2023-1234", "https://nvd.nist.gov/vuln/detail/CVE-2023-1234"]),
        ("unknown-format", None),
    ],
)
def test_build_reference_links(vulnerability_id: str, expected: list[str] | None):
    assert build_reference_links(vulnerability_id) == expected



from __future__ import annotations

import logging
import os
import shlex
import shutil
import subprocess
import tempfile
from dataclasses import dataclass

from vunnel import utils


@dataclass
class GitRevision:
    sha: str
    file: str


class GitWrapper:
    _check_cmd_ = "git --version"
    _is_git_repo_cmd_ = "git rev-parse --is-inside-work-tree"
    _clone_cmd_ = "git clone -b {branch} {src} {dest}"
    _check_out_cmd_ = "git checkout {branch}"

    def __init__(
        self,
        source: str,
        branch: str,
        checkout_dest: str,
        logger: logging.Logger | None = None,
    ):
        self.src = source
        self.branch = branch
        self.dest = checkout_dest
        self.workspace = tempfile.gettempdir()

        if not logger:
            logger = logging.getLogger(self.__class__.__name__)
        self.logger = logger

        try:
            out = self._exec_cmd(self._check_cmd_)
            self.logger.trace(f"git executable verified using cmd: {self._check_cmd_}, output: {out}")  # type: ignore[attr-defined]
        except:
            self.logger.exception('could not find required "git" executable. Please install git on host')
            raise

    def delete_repo(self) -> None:
        if os.path.exists(self.dest):
            self.logger.debug("deleting existing repository")
            shutil.rmtree(self.dest, ignore_errors=True)

    @utils.retry_with_backoff()
    def clone_repo(self) -> None:
        try:
            self.logger.info(f"cloning git repository {self.src} branch {self.branch} to {self.dest}")
            cmd = self._clone_cmd_.format(src=self.src, dest=self.dest, branch=self.branch)
            out = self._exec_cmd(cmd)
            self.logger.debug(f"initialized git repo, cmd: {cmd}, output: {out}")
        except:
            self.logger.exception(f"failed to clone git repository {self.src} branch {self.branch} to {self.dest}")
            raise

    def _exec_cmd(self, cmd: str) -> str:
        """
        Run a command with errors etc handled
        :param cmd: list of arguments (including command name, e.g. ['ls', '-l])
        :param args:
        :param kwargs:
        :return:
        """
        try:
            self.logger.trace(f"running: {cmd}")  # type: ignore[attr-defined]
            cmd_list = shlex.split(cmd)
            # S603 disable explanation: running git commands by design
            return subprocess.check_output(cmd_list, text=True, stderr=subprocess.PIPE)  # noqa: S603
        except Exception as e:
            self.logger.exception(f"error executing command: {cmd}")
            raise e



from collections.abc import Generator as IterGenerator
from dataclasses import dataclass, field

from mashumaro import field_options
from mashumaro.config import BaseConfig
from mashumaro.mixins.orjson import DataClassORJSONMixin


class OmitNoneORJSONModel(DataClassORJSONMixin):
    class Config(BaseConfig):
        omit_none = True


@dataclass
class CVSS_V3(DataClassORJSONMixin):
    attack_complexity: str = field(metadata=field_options(alias="attackComplexity"))
    attack_vector: str = field(metadata=field_options(alias="attackVector"))
    availability_impact: str = field(metadata=field_options(alias="availabilityImpact"))
    base_score: float = field(metadata=field_options(alias="baseScore"))
    base_severity: str = field(metadata=field_options(alias="baseSeverity"))
    confidentiality_impact: str = field(metadata=field_options(alias="confidentialityImpact"))
    integrity_impact: str = field(metadata=field_options(alias="integrityImpact"))
    privileges_required: str = field(metadata=field_options(alias="privilegesRequired"))
    scope: str = field(metadata=field_options(alias="scope"))
    user_interaction: str = field(metadata=field_options(alias="userInteraction"))
    vector_string: str = field(metadata=field_options(alias="vectorString"))
    version: str = field(metadata=field_options(alias="version"))

    class Config(BaseConfig):
        serialize_by_alias = True  # normal CSAF is snake_case, but embeds camelCase CVSS objects


@dataclass
class CVSS_V2(DataClassORJSONMixin):
    access_complexity: str = field(metadata=field_options(alias="accessComplexity"))
    access_vector: str = field(metadata=field_options(alias="accessVector"))
    authentication: str = field(metadata=field_options(alias="authentication"))
    availability_impact: str = field(metadata=field_options(alias="availabilityImpact"))
    base_score: float = field(metadata=field_options(alias="baseScore"))
    confidentiality_impact: str = field(metadata=field_options(alias="confidentialityImpact"))
    integrity_impact: str = field(metadata=field_options(alias="integrityImpact"))
    vector_string: str = field(metadata=field_options(alias="vectorString"))
    version: str = field(metadata=field_options(alias="version"))

    class Config(BaseConfig):
        serialize_by_alias = True  # normal CSAF is snake_case, but embeds camelCase CVSS objects


@dataclass
class Reference(OmitNoneORJSONModel):
    category: str
    summary: str
    url: str


@dataclass
class Note(OmitNoneORJSONModel):
    category: str
    text: str
    title: str


@dataclass
class ProductStatus(OmitNoneORJSONModel):
    fixed: list[str] = field(default_factory=list)
    known_affected: list[str] = field(default_factory=list)
    known_not_affected: list[str] = field(default_factory=list)
    under_investigation: list[str] = field(default_factory=list)


@dataclass
class Threat(OmitNoneORJSONModel):
    category: str
    details: str
    product_ids: list[str] = field(default_factory=list)


@dataclass
class CWE(OmitNoneORJSONModel):
    id: str
    name: str


@dataclass
class Flag(OmitNoneORJSONModel):
    label: str
    product_ids: list[str]


@dataclass
class VulnID(OmitNoneORJSONModel):
    system_name: str
    text: str


@dataclass
class Remediation(OmitNoneORJSONModel):
    category: str
    details: str
    product_ids: list[str]
    url: str | None = None


@dataclass
class Score(OmitNoneORJSONModel):
    products: list[str]
    cvss_v3: CVSS_V3 | None = None
    cvss_v2: CVSS_V2 | None = None


@dataclass
class Vulnerability(OmitNoneORJSONModel):
    title: str
    cve: str
    cwe: str | None = None
    discovery_date: str | None = None
    flags: list[Flag] = field(default_factory=list)
    ids: list[VulnID] = field(default_factory=list)
    notes: list[Note] = field(default_factory=list)
    product_status: ProductStatus | None = None
    references: list[Reference] | None = None
    release_date: str | None = None
    remediations: list[Remediation] = field(default_factory=list)
    scores: list[Score] = field(default_factory=list)
    threats: list[Threat] = field(default_factory=list)


@dataclass
class FullProductName(OmitNoneORJSONModel):
    name: str
    product_id: str


@dataclass
class Relationship(OmitNoneORJSONModel):
    category: str
    full_product_name: FullProductName
    product_reference: str
    relates_to_product_reference: str


@dataclass
class ProductIdentificationHelper(OmitNoneORJSONModel):
    cpe: str | None = None
    purl: str | None = None


@dataclass
class Product(OmitNoneORJSONModel):
    name: str
    product_id: str
    product_identification_helper: ProductIdentificationHelper | None = None


@dataclass
class Branch(OmitNoneORJSONModel):
    category: str
    name: str
    branches: list["Branch"] = field(default_factory=list)
    product: Product | None = None

    def purl(self) -> str | None:
        if self.product and self.product.product_identification_helper:
            return self.product.product_identification_helper.purl
        return None

    def cpe(self) -> str | None:
        if self.product and self.product.product_identification_helper:
            return self.product.product_identification_helper.cpe
        return None

    def product_id(self) -> str | None:
        if self.product:
            return self.product.product_id
        return None

    def product_branches(self) -> IterGenerator["Branch", None, None]:
        yield self
        for b in self.branches:
            yield from b.product_branches()

    def product_version_branches(self) -> IterGenerator["Branch", None, None]:
        for b in self.product_branches():
            if b.category == "product_version":
                yield b

    def product_name_branches(self) -> IterGenerator["Branch", None, None]:
        for b in self.product_branches():
            if b.category == "product_name":
                yield b


@dataclass
class ProductTree(OmitNoneORJSONModel):
    relationships: list[Relationship] = field(default_factory=list)
    branches: list[Branch] = field(default_factory=list)
    product_id_to_parent: dict[str, str] = field(
        init=False,
        metadata={"serialize": "omit"},  # field is a cache for runtime efficiency, but not part of spec
    )
    product_id_to_purl: dict[str, str] = field(
        init=False,
        metadata={"serialize": "omit"},  # field is a cache for runtime efficiency, but not part of spec
    )

    def __post_init__(self) -> None:
        self.product_id_to_parent = {}
        for r in self.relationships:
            self.product_id_to_parent[r.full_product_name.product_id] = r.relates_to_product_reference

        self.product_id_to_purl = {}
        for b in self.product_branches():
            purl = b.purl()
            pid = b.product_id()
            if purl and pid:
                self.product_id_to_purl[pid] = purl

    def product_branches(self) -> IterGenerator[Branch, None, None]:
        for b in self.branches:
            yield from b.product_branches()

    def parent(self, product_id: str) -> str | None:
        return self.product_id_to_parent.get(product_id)

    def purl_for_product_id(self, product_id: str) -> str | None:
        return self.product_id_to_purl.get(product_id)


@dataclass
class AggregateSeverity(OmitNoneORJSONModel):
    namespace: str
    text: str


@dataclass
class TLP(OmitNoneORJSONModel):
    label: str
    url: str


@dataclass
class Distribution(OmitNoneORJSONModel):
    text: str
    tlp: TLP


@dataclass
class Publisher(OmitNoneORJSONModel):
    category: str
    contact_details: str
    issuing_authority: str
    name: str
    namespace: str


@dataclass
class GeneratorEngine(OmitNoneORJSONModel):
    name: str
    version: str


@dataclass
class Generator(OmitNoneORJSONModel):
    date: str
    engine: GeneratorEngine


@dataclass
class RevisionEntry(OmitNoneORJSONModel):
    date: str
    number: str  # yes, really
    summary: str


@dataclass
class Tracking(OmitNoneORJSONModel):
    current_release_date: str
    generator: Generator
    id: str
    initial_release_date: str
    revision_history: list[RevisionEntry]
    status: str
    version: str


@dataclass
class Document(OmitNoneORJSONModel):
    aggregate_severity: AggregateSeverity
    category: str
    csaf_version: str
    distribution: Distribution
    lang: str
    notes: list[Note]
    publisher: Publisher
    references: list[Reference]
    title: str
    tracking: Tracking


@dataclass
class CSAFDoc(OmitNoneORJSONModel):
    document: Document
    product_tree: ProductTree
    vulnerabilities: list[Vulnerability] = field(default_factory=list)


def from_path(path: str) -> CSAFDoc:
    with open(path) as fh:
        return CSAFDoc.from_json(fh.read())



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any

from lxml import etree
from xsdata.formats.dataclass.parsers import XmlParser
from xsdata.formats.dataclass.parsers.config import ParserConfig

from vunnel.providers.mariner.model import Definition, RpminfoObject, RpminfoState, RpminfoTest
from vunnel.utils import http_wrapper as http
from vunnel.utils.vulnerability import FixedIn, Vulnerability

if TYPE_CHECKING:
    import logging
    from collections.abc import Generator

    from vunnel.workspace import Workspace

LESS_THAN_OR_EQUAL_TO = "less than or equal"
LESS_THAN = "less than"
GREATER_THAN = "greater than"

IGNORED_PATCHABLE_VALUES = ["Not Applicable"]


class MarinerXmlFile:
    def __init__(self, oval_file_path: str, logger: logging.Logger):
        parser_config = ParserConfig(
            fail_on_converter_warnings=False,
            fail_on_unknown_attributes=False,
            fail_on_unknown_properties=False,
        )
        xml_parser = XmlParser(config=parser_config)
        # S320 disable explanation: the mariner linux vulnerability feed is not untrusted xml
        root = etree.parse(oval_file_path)  # noqa: S320
        nsmap = etree.XPath("/*")(root)[0].nsmap
        default = nsmap[None]
        nsmap["default"] = default
        del nsmap[None]
        self.logger = logger

        # Go element by element to have flexible error handling
        self.definitions = []
        for def_element in etree.XPath("//default:definition", namespaces=nsmap)(root):
            try:
                element_bytes = etree.tostring(def_element)
                definition = xml_parser.from_string(element_bytes.decode(), Definition)
                self.definitions.append(definition)
            except Exception as ex:
                self.logger.warning(f"skipping definition element in {oval_file_path} due to {ex}")

        self.tests_by_id = {}
        for test_element in etree.XPath("//linux-def:rpminfo_test", namespaces=nsmap)(root):
            try:
                element_bytes = etree.tostring(test_element)
                test_obj = xml_parser.from_string(element_bytes.decode(), RpminfoTest)
                self.tests_by_id[test_obj.id] = test_obj
            except Exception as ex:
                self.logger.warning(f"skipping rpminfo_test element in {oval_file_path} due to {ex}")
        self.objects_by_id = {}
        for obj_element in etree.XPath("//linux-def:rpminfo_object", namespaces=nsmap)(root):
            try:
                element_bytes = etree.tostring(obj_element)
                obj = xml_parser.from_string(element_bytes.decode(), RpminfoObject)
                self.objects_by_id[obj.id] = obj
            except Exception as ex:
                self.logger.warning(f"skipping rpminfo_object element in {oval_file_path} due to {ex}")

        self.states_by_id = {}
        for state_element in etree.XPath("//linux-def:rpminfo_state", namespaces=nsmap)(root):
            try:
                element_bytes = etree.tostring(state_element)
                state_obj = xml_parser.from_string(element_bytes.decode(), RpminfoState)
                self.states_by_id[state_obj.id] = state_obj
            except Exception as ex:
                self.logger.warning(f"skipping rpminfo_object element in {oval_file_path} due to {ex}")
        self.mariner_version = oval_file_path.split("-")[-2]

    def name_and_version(self, test_id: str) -> tuple[str | None, str | None]:
        test = self.tests_by_id.get(test_id, None)
        if test is None or test.object_value is None or test.state is None:
            return (None, None)
        obj = self.objects_by_id.get(test.object_value.object_ref, None)
        if obj is None:
            return (None, None)
        state = self.states_by_id.get(test.state.state_ref, None)
        if state is None or state.evr is None:
            return (None, None)
        return (obj.name, state.evr.value)

    def namespace_name(self) -> str:
        return f"mariner:{self.mariner_version}"

    def get_tests(self, definition: Definition) -> list[RpminfoTest]:
        tests = []
        if definition and definition.criteria and definition.criteria.criterion:
            for criterion in definition.criteria.criterion:
                test = self.tests_by_id.get(criterion.test_ref, None)
                if test:
                    tests.append(test)
        return tests

    def get_states(self, tests: list[RpminfoTest]) -> list[RpminfoState]:
        states = []
        for test in tests:
            if test and test.state and test.state.state_ref:
                state = self.states_by_id.get(test.state.state_ref, None)
                if state:
                    states.append(state)
        return states

    def get_objects(self, tests: list[RpminfoTest]) -> list[RpminfoObject]:
        objects = []
        for test in tests:
            if test and test.object_value and test.object_value.object_ref:
                obj = self.objects_by_id.get(test.object_value.object_ref, None)
                if obj:
                    objects.append(obj)
        return objects

    def make_fixed_in(self, definition: Definition) -> FixedIn | None:
        tests = self.get_tests(definition)
        states = self.get_states(tests)
        objects = self.get_objects(tests)

        if not states or not objects:
            return None

        name = objects[0].name
        if not name:
            return None

        fixed_version = None

        vulnerability_range = []
        for state in states:
            if state.evr and state.evr.value:
                if state.evr.operation == LESS_THAN:
                    vulnerability_range.append(f"< {state.evr.value}")
                    # if vulnerability has an upper bound (< as opposed to <=),
                    # then assume the upper bound is the fixed version
                    fixed_version = state.evr.value
                elif state.evr.operation == GREATER_THAN:
                    vulnerability_range.append(f"> {state.evr.value}")
                elif state.evr.operation == LESS_THAN_OR_EQUAL_TO:
                    vulnerability_range.append(f"<= {state.evr.value}")

        if not vulnerability_range:
            return None

        # make output deterministic. Reverse so that
        # output reads like >1.2.3 <1.3.0 instead of the reverse.
        vulnerability_range.sort(reverse=True)

        if not fixed_version:
            fixed_version = "None"  # a required string in JSON schema

        vulnerability_range_str = ", ".join(vulnerability_range)

        return FixedIn(
            Name=name,
            NamespaceName=self.namespace_name(),
            VersionFormat="rpm",
            Version=fixed_version,
            VulnerableRange=vulnerability_range_str,
            Module=None,
            VendorAdvisory=None,
        )

    def vulnerability_id(self, definition: Definition) -> str | None:
        if definition.metadata is None or definition.metadata.reference is None or definition.metadata.reference.ref_id is None:
            return None
        return definition.metadata.reference.ref_id

    def description(self, definition: Definition) -> str | None:
        if not definition.metadata:
            return None
        return definition.metadata.description

    def vulnerabilities(self) -> Generator[Vulnerability, None, None]:
        for d in self.definitions:
            if d.metadata is None or d.metadata.severity is None:
                self.logger.warning("skipping definition because severity could not be found")
                continue
            if d.metadata and d.metadata.patchable and d.metadata.patchable in IGNORED_PATCHABLE_VALUES:
                continue
            link = ""
            if d.metadata.reference and d.metadata.reference.ref_url:
                link = d.metadata.reference.ref_url
            fixed_in = self.make_fixed_in(d)
            if not fixed_in:
                continue
            vulnerability_id = self.vulnerability_id(d)
            if not vulnerability_id:
                continue
            yield Vulnerability(
                Name=vulnerability_id,  # intentional; legacy API uses Name as field for vulnerability ID.
                NamespaceName=self.namespace_name(),
                Description=self.description(d) or "",
                Severity=d.metadata.severity,
                Link=link,
                CVSS=[],
                FixedIn=[fixed_in],
                Metadata={},  # Ideally, there's no metadata here.
            )


MARINER_URL_BASE = "https://raw.githubusercontent.com/microsoft/CBL-MarinerVulnerabilityData/main/{}"
MARINER_URL_FILENAME = "cbl-mariner-{}-oval.xml"
AL3_URL = "https://raw.githubusercontent.com/microsoft/AzureLinuxVulnerabilityData/main/azurelinux-3.0-oval.xml"

VERSION_TO_URL = {
    "1.0": MARINER_URL_BASE.format(MARINER_URL_FILENAME.format("1.0")),
    "2.0": MARINER_URL_BASE.format(MARINER_URL_FILENAME.format("2.0")),
    "3.0": AL3_URL,
}

VERSION_TO_FILENAME = {
    "1.0": MARINER_URL_FILENAME.format("1.0"),
    "2.0": MARINER_URL_FILENAME.format("2.0"),
    "3.0": "azurelinux-3.0-oval.xml",
}


class Parser:
    def __init__(self, workspace: Workspace, download_timeout: int, allow_versions: list[Any], logger: logging.Logger):
        self.workspace = workspace
        self.download_timeout = download_timeout
        self.allow_versions = allow_versions
        self._urls: set[str] = set()
        self.logger = logger

    def _download(self) -> list[str]:
        return [self._download_version(v) for v in self.allow_versions]

    def _download_version(self, version: str) -> str:
        filename = VERSION_TO_FILENAME[version]
        if not filename:
            raise Exception(f"mariner/azurelinux provider misconfigured: no filename for version {version}")
        url = VERSION_TO_URL[version]
        if not url:
            raise Exception(f"mariner/azurelinux provider misconfigured: no URL for version {version}")
        r = http.get(url, self.logger, timeout=self.download_timeout)
        destination = os.path.join(self.workspace.input_path, filename)
        with open(destination, "wb") as writer:
            writer.write(r.content)
        self._urls.add(url)
        return destination

    @property
    def urls(self) -> list[str]:
        return list(self._urls)

    def get(self) -> Generator[tuple[str, str, dict[str, dict[str, Any]]], None, None]:
        for oval_file_path in self._download():
            parsed_file = MarinerXmlFile(oval_file_path, self.logger)
            for v in parsed_file.vulnerabilities():
                yield v.NamespaceName, v.Name, v.to_payload()



from __future__ import annotations

import enum
import re
import tempfile

import defusedxml.ElementTree as ET
import pytest
from vunnel.utils.oval_v2 import (
    Artifact,
    ArtifactParser,
    OVALElementEnum,
    OVALElementParser,
    OVALParserConfig,
    OVALParserFactory,
    Test,
    TestParser,
    Version,
    VersionParser,
    iter_parse_vulnerability_file,
)

# in order to avoid pytest to collect this class as a test
# (e.g. "PytestCollectionWarning: cannot collect test class 'Test' because it has a __init__ constructor")
Test.__test__ = False
TestParser.__test__ = False


@pytest.fixture()
def dummy_config():
    return OVALParserConfig(
        platform_regex=None,
        artifact_regex=None,
        source_url_xpath_query=None,
        severity_xpath_query=None,
        severity_map=None,
    )


class TestTestParser:
    @pytest.mark.parametrize(
        "element",
        [
            pytest.param(
                '<rpminfo_test id="oval:org.opensuse.security:tst:2009223735" version="1" comment="sle-module-basesystem-release is ==15" check="at least one" xmlns="http://oval.mitre.org/XMLSchema/oval-definitions-5#linux">'  # noqa: E501
                + '<object object_ref="oval:org.opensuse.security:obj:2009042619"/>'
                + '<state state_ref="oval:org.opensuse.security:ste:2009061809"/>'
                + "</rpminfo_test>",
                id="with-namespace",
            ),
            pytest.param(
                '<rpminfo_test id="oval:org.opensuse.security:tst:2009223735" version="1" comment="sle-module-basesystem-release is ==15" check="at least one">'  # noqa: E501
                + '<object object_ref="oval:org.opensuse.security:obj:2009042619"/>'
                + '<state state_ref="oval:org.opensuse.security:ste:2009061809"/>'
                + "</rpminfo_test>",
                id="without-namespace",
            ),
        ],
    )
    def test_happy_path(self, dummy_config, element):
        xml_element = ET.fromstring(element)
        result = TestParser.parse(xml_element, dummy_config)
        assert result
        assert isinstance(result, Test)
        assert result.identity == "oval:org.opensuse.security:tst:2009223735"
        assert result.artifact_id == "oval:org.opensuse.security:obj:2009042619"
        assert result.version_id == "oval:org.opensuse.security:ste:2009061809"

    @pytest.mark.parametrize(
        "element",
        [
            pytest.param(None, id="none"),
            pytest.param("", id="blank"),
            pytest.param(ET.fromstring('<rpminfo_object id="dummy"/>'), id="not-test"),
            pytest.param(
                ET.fromstring('<rpminfo_test id="dummy" xmlns="http://oval.mitre.org/XMLSchema/oval-definitions-5#linux"/>'),
                id="invalid-test",
            ),
        ],
    )
    def test_unhappy_paths(self, dummy_config, element):
        assert TestParser.parse(element, dummy_config) is None


class TestArtifactParser:
    @pytest.mark.parametrize(
        "element",
        [
            pytest.param(
                '<rpminfo_object id="oval:org.opensuse.security:obj:2009041419" version="1" xmlns="http://oval.mitre.org/XMLSchema/oval-definitions-5#linux">'  # noqa: E501
                + "<name>policycoreutils-python</name>"
                + "</rpminfo_object>",
                id="with-namespace",
            ),
            pytest.param(
                '<rpminfo_object id="oval:org.opensuse.security:obj:2009041419" version="1">'
                + "<name>policycoreutils-python</name>"
                + "</rpminfo_object>",
                id="without-namespace",
            ),
        ],
    )
    def test_happy_path(self, dummy_config, element):
        xml_element = ET.fromstring(element)
        result = ArtifactParser.parse(xml_element, dummy_config)
        assert result
        assert isinstance(result, Artifact)
        assert result.identity == "oval:org.opensuse.security:obj:2009041419"
        assert result.name == "policycoreutils-python"

    @pytest.mark.parametrize(
        "element",
        [
            pytest.param(None, id="none"),
            pytest.param("", id="blank"),
            pytest.param(ET.fromstring('<rpminfo_state id="dummy"/>'), id="not-artifact"),
            pytest.param(
                ET.fromstring(
                    '<rpminfo_object id="dummy" version="1" xmlns="http://oval.mitre.org/XMLSchema/oval-definitions-5#linux"/>',  # noqa: E501
                ),
                id="invalid-artifact",
            ),
        ],
    )
    def test_unhappy_paths(self, dummy_config, element):
        assert ArtifactParser.parse(element, dummy_config) is None


class TestVersionParser:
    @pytest.mark.parametrize(
        ("element", "identity", "op", "value"),
        [
            pytest.param(
                '<rpminfo_state id="oval:org.opensuse.security:ste:2009079458" version="1" xmlns="http://oval.mitre.org/XMLSchema/oval-definitions-5#linux">'  # noqa: E501
                + '<version operation="equals">0</version>'
                + "</rpminfo_state>",
                "oval:org.opensuse.security:ste:2009079458",
                "equals",
                "0",
                id="version-with-namespace",
            ),
            pytest.param(
                '<rpminfo_state id="oval:org.opensuse.security:ste:2009111500" version="1" xmlns="http://oval.mitre.org/XMLSchema/oval-definitions-5#linux">'  # noqa: E501
                + '<evr datatype="evr_string" operation="less than">0:1.15.2-4.25</evr>'
                + '<arch datatype="string" operation="pattern match">(noarch)</arch>'
                + "</rpminfo_state>",
                "oval:org.opensuse.security:ste:2009111500",
                "less than",
                "0:1.15.2-4.25",
                id="evr-with-namespace",
            ),
            pytest.param(
                '<rpminfo_state id="oval:org.opensuse.security:ste:2009079458" version="1">'
                + '<version operation="equals">0</version>'
                + "</rpminfo_state>",
                "oval:org.opensuse.security:ste:2009079458",
                "equals",
                "0",
                id="version-without-namespace",
            ),
            pytest.param(
                '<rpminfo_state id="oval:org.opensuse.security:ste:2009111500" version="1">'
                + '<evr datatype="evr_string" operation="less than">0:1.15.2-4.25</evr>'
                + "</rpminfo_state>",
                "oval:org.opensuse.security:ste:2009111500",
                "less than",
                "0:1.15.2-4.25",
                id="evr-without-namespace",
            ),
        ],
    )
    def test_happy_path(self, dummy_config, element, identity, op, value):
        xml_element = ET.fromstring(element)
        result = VersionParser.parse(xml_element, dummy_config)
        assert result
        assert isinstance(result, Version)
        assert result.identity == identity
        assert result.operation == op
        assert result.value == value

    @pytest.mark.parametrize(
        "element",
        [
            pytest.param(None, id="none"),
            pytest.param("", id="blank"),
            pytest.param(ET.fromstring('<rpminfo_test id="dummy"/>'), id="not-version"),
            pytest.param(
                ET.fromstring(
                    '<rpminfo_state id="dummy" version="1" xmlns="http://oval.mitre.org/XMLSchema/oval-definitions-5#linux"/>',  # noqa: E501
                ),
                id="invalid-version",
            ),
        ],
    )
    def test_unhappy_paths(self, dummy_config, element):
        assert ArtifactParser.parse(element, dummy_config) is None


class TestParserFactory:
    @pytest.fixture()
    def invalid_parsers_error(self):
        return "Invalid input for parsers, must be a list of OVALElementParser sub-classes"

    @pytest.fixture()
    def invalid_enum_error(self):
        return "Invalid input for oval element enumeration, must be a python enum class"

    @pytest.fixture()
    def parser_enum_mismatch_error(self):
        return "Parsers are not a match for the oval element enumeration"

    @pytest.fixture()
    def example_enum_class(self):
        class RandomEnum(enum.Enum):
            FOO = "foo"

        return RandomEnum

    @pytest.fixture()
    def example_parser_class(self, example_enum_class):
        class RandomParser(OVALElementParser):
            oval_element = example_enum_class.FOO

            @staticmethod
            def parse(xml_element, config: OVALParserConfig):
                pass

        return RandomParser

    @pytest.mark.parametrize(
        ("parsers", "oval_enum"),
        [
            pytest.param(
                [
                    TestParser,
                    ArtifactParser,
                    VersionParser,
                ],
                OVALElementEnum,
                id="missing-parser",
            ),
        ],
    )
    def test_init_parser_enum_mismatch(self, parsers, oval_enum, parser_enum_mismatch_error):
        with pytest.raises(ValueError) as e:
            OVALParserFactory(parsers, oval_enum)
        assert str(e.value) == parser_enum_mismatch_error

    @pytest.mark.parametrize(
        "parsers",
        [
            pytest.param(None, id="none"),
            pytest.param({}, id="wrong-type-dict"),
            pytest.param([], id="empty"),
            pytest.param("", id="wrong-type-string"),
        ],
    )
    def test_init_invalid_parsers(self, parsers, invalid_parsers_error):
        with pytest.raises(ValueError) as e:
            OVALParserFactory(parsers, None)
        assert str(e.value) == invalid_parsers_error

    @pytest.mark.parametrize(
        "test_enum_class",
        [
            pytest.param(None, id="none"),
            pytest.param({}, id="wrong-type-dict"),
            pytest.param("", id="wrong-type-string"),
        ],
    )
    def test_init_invalid_oval_enum(self, test_enum_class, invalid_enum_error):
        with pytest.raises(ValueError) as e:
            OVALParserFactory([TestParser], test_enum_class)
        assert str(e.value) == invalid_enum_error

    def test_init_happy_path(self, example_enum_class, example_parser_class):
        pf = OVALParserFactory(
            parsers=[example_parser_class],
            element_enum=example_enum_class,
        )
        assert pf
        assert isinstance(pf, OVALParserFactory)

    def test_get_parser_happy_path(self, example_enum_class, example_parser_class):
        pf = OVALParserFactory(
            parsers=[example_parser_class],
            element_enum=example_enum_class,
        )
        parser = pf.get_parser(example_enum_class.FOO)
        assert parser
        assert parser == example_parser_class

        pf.parser_map = {}
        parser = pf.get_parser(example_enum_class.FOO)
        assert parser is None

    @pytest.mark.parametrize(
        "test_enum_class",
        [
            pytest.param(None, id="none"),
            pytest.param({}, id="wrong-type-dict"),
            pytest.param("", id="wrong-type-string"),
            pytest.param(OVALElementEnum.TEST, id="wrong-type-enum"),
        ],
    )
    def test_get_parser_invalid_input(self, test_enum_class, example_enum_class, example_parser_class):
        pf = OVALParserFactory(
            parsers=[example_parser_class],
            element_enum=example_enum_class,
        )
        parser = pf.get_parser(test_enum_class)

        assert parser is None

    @pytest.mark.parametrize(
        "test_xml_element",
        [
            pytest.param(ET.fromstring('<foo id="dummy"/>'), id="valid-tag"),
            pytest.param(
                ET.fromstring('<foo id="dummy" version="1" xmlns="http://oval.mitre.org/XMLSchema/oval-definitions-5#linux"/>'),
                id="valid-tag-with-ns",
            ),
        ],
    )
    def test_get_oval_element_happy_path(
        self,
        test_xml_element,
        example_enum_class,
        example_parser_class,
        dummy_config,
    ):
        pf = OVALParserFactory(
            parsers=[example_parser_class],
            element_enum=example_enum_class,
        )

        assert pf.get_oval_element(test_xml_element, dummy_config) == example_enum_class.FOO

    @pytest.mark.parametrize(
        "test_xml_element",
        [
            pytest.param(None, id="none"),
            pytest.param("", id="blank"),
            pytest.param("foobar", id="wrong-type-string"),
            pytest.param(ET.fromstring('<rpminfo_state id="dummy"/>'), id="invalid-tag"),
            pytest.param(
                ET.fromstring(
                    '<rpminfo_object id="dummy" version="1" xmlns="http://oval.mitre.org/XMLSchema/oval-definitions-5#linux"/>',
                ),
                id="invalid-tag-with-ns",
            ),
        ],
    )
    def test_get_oval_element_invalid_input(
        self,
        test_xml_element,
        example_enum_class,
        example_parser_class,
        dummy_config,
    ):
        pf = OVALParserFactory(
            parsers=[example_parser_class],
            element_enum=example_enum_class,
        )

        assert pf.get_oval_element(test_xml_element, dummy_config) is None


class TestOVALElementParser:
    @pytest.mark.parametrize(
        ("data", "regex", "expected"),
        [
            pytest.param(
                "{http://oval.mitre.org/XMLSchema/oval-definitions-5#linux}object",
                re.compile(r"\{http://oval.mitre.org/XMLSchema/.*\}(\w*)"),
                "object",
                id="with-namespace",
            ),
            pytest.param(
                "object",
                re.compile(r"\{http://oval.mitre.org/XMLSchema/.*\}(\w*)"),
                "object",
                id="without-namespace",
            ),
        ],
    )
    def test_find_with_regex_happy_path(self, data, regex, expected):
        assert OVALElementParser._find_with_regex(data, regex) == expected

    @pytest.mark.parametrize(
        ("data", "regex"),
        [
            pytest.param(
                None,
                None,
                id="invalid-1",
            ),
            pytest.param(
                None,
                re.compile(r"foo"),
                id="invalid-2",
            ),
            pytest.param(
                "",
                None,
                id="invalid-3",
            ),
        ],
    )
    def test_find_with_regex_invalid_input(self, data, regex):
        with pytest.raises(Exception):
            OVALElementParser._find_with_regex(data, regex)


class TestIterParse:
    @pytest.mark.parametrize(
        ("content", "tag", "parser_fn", "expected"),
        [
            pytest.param(
                '<rpminfo_test id="oval:org.opensuse.security:tst:2009223735" version="1" comment="sle-module-basesystem-release is ==15" check="at least one" xmlns="http://oval.mitre.org/XMLSchema/oval-definitions-5#linux">'  # noqa: E501
                + '<object object_ref="oval:org.opensuse.security:obj:2009042619"/>'
                + '<state state_ref="oval:org.opensuse.security:ste:2009061809"/>'
                + "</rpminfo_test>",
                "rpminfo_test",
                TestParser.parse,
                Test(
                    identity="oval:org.opensuse.security:tst:2009223735",
                    artifact_id="oval:org.opensuse.security:obj:2009042619",
                    version_id="oval:org.opensuse.security:ste:2009061809",
                ),
                id="test-with-namespace",
            ),
            pytest.param(
                '<rpminfo_object id="oval:org.opensuse.security:obj:2009041419" version="1">'
                + "<name>policycoreutils-python</name>"
                + "</rpminfo_object>",
                "rpminfo_object",
                ArtifactParser.parse,
                Artifact(
                    identity="oval:org.opensuse.security:obj:2009041419",
                    name="policycoreutils-python",
                ),
                id="artifact-without-namespace",
            ),
        ],
    )
    def test_iter_parse_vulnerability_file_happy_path(self, dummy_config, content, tag, parser_fn, expected):
        class RandomElements(enum.Enum):
            FOO = tag

        class RandomParser(OVALElementParser):
            oval_element = RandomElements.FOO
            parse = parser_fn

        with tempfile.NamedTemporaryFile(mode="w") as fp:
            fp.write(content)
            fp.seek(0)

            results = iter_parse_vulnerability_file(
                fp.name,
                parser_config=dummy_config,
                parser_factory=OVALParserFactory(parsers=[RandomParser], element_enum=RandomElements),
            )

        assert results
        assert isinstance(results, dict)
        assert len(results) == 1
        assert list(results.keys())[0] == RandomElements.FOO

        foo_map = results.get(RandomElements.FOO)
        assert len(foo_map) == 1
        assert list(foo_map.keys())[0] == expected.identity
        assert foo_map.get(expected.identity) == expected



from vunnel import schema as schema_def


def test_provider_workspace_schema_v1():
    # it is vital that we do not make any breaking changes to the provider workspace state schema
    # until there is a mechanism to deal with the state version detection, migration, and possibly supporting
    # multiple version implementations in the codebase
    assert schema_def.PROVIDER_WORKSPACE_STATE_SCHEMA_VERSION.startswith("1.")
    assert schema_def.ProviderStateSchema().version.startswith("1.")



from __future__ import annotations

import shutil

import pytest
from pytest_unordered import unordered
from vunnel import result, workspace
from vunnel.providers.oracle import Config, Provider, parser


@pytest.mark.parametrize(
    ("input_file", "expected"),
    [
        # TEST 1 ############################################################
        (
            "test-fixtures/mock_ksplice_data",
            {
                ("ELSA-2021-9151", "ol:8"): (
                    "501",
                    {
                        "Vulnerability": {
                            "Severity": "High",
                            "NamespaceName": "ol:8",
                            "FixedIn": [],
                            "Link": "https://linux.oracle.com/errata/ELSA-2021-9151.html",
                            "Description": "[1.1.1g-15] - version bump [1.1.1g-14] - CVE-2021-3450 openssl: CA certificate check bypass with X509_V_FLAG_X509_STRICT [1.1.1g-13] - Fix CVE-2021-3449 NULL pointer deref in signature_algorithms processing",
                            "Metadata": {
                                "Issued": "2021-04-01",
                                "RefId": "ELSA-2021-9151",
                                "CVE": [
                                    {
                                        "Name": "CVE-2021-3449",
                                        "Link": "https://linux.oracle.com/cve/CVE-2021-3449.html",
                                    },
                                    {
                                        "Name": "CVE-2021-3450",
                                        "Link": "https://linux.oracle.com/cve/CVE-2021-3450.html",
                                    },
                                ],
                            },
                            "Name": "ELSA-2021-9151",
                            "CVSS": [],
                        },
                    },
                ),
            },
        ),
        # TEST 2 ############################################################
        (
            "test-fixtures/mock_data",
            {
                ("ELSA-2007-0057", "ol:5"): (
                    "501",
                    {
                        "Vulnerability": {
                            "Severity": "Medium",
                            "NamespaceName": "ol:5",
                            "FixedIn": unordered(
                                [
                                    {
                                        "Name": "bind-utils",
                                        "Version": "30:9.3.3-8.el5",
                                        "VersionFormat": "rpm",
                                        "NamespaceName": "ol:5",
                                        "Module": None,
                                    },
                                    {
                                        "Name": "bind-devel",
                                        "Version": "30:9.3.3-8.el5",
                                        "VersionFormat": "rpm",
                                        "NamespaceName": "ol:5",
                                        "Module": None,
                                    },
                                    {
                                        "Name": "bind-sdb",
                                        "Version": "30:9.3.3-8.el5",
                                        "VersionFormat": "rpm",
                                        "NamespaceName": "ol:5",
                                        "Module": None,
                                    },
                                    {
                                        "Name": "bind-libbind-devel",
                                        "Version": "30:9.3.3-8.el5",
                                        "VersionFormat": "rpm",
                                        "NamespaceName": "ol:5",
                                        "Module": None,
                                    },
                                    {
                                        "Name": "bind",
                                        "Version": "30:9.3.3-8.el5",
                                        "VersionFormat": "rpm",
                                        "NamespaceName": "ol:5",
                                        "Module": None,
                                    },
                                    {
                                        "Name": "caching-nameserver",
                                        "Version": "30:9.3.3-8.el5",
                                        "VersionFormat": "rpm",
                                        "NamespaceName": "ol:5",
                                        "Module": None,
                                    },
                                    {
                                        "Name": "bind-chroot",
                                        "Version": "30:9.3.3-8.el5",
                                        "VersionFormat": "rpm",
                                        "NamespaceName": "ol:5",
                                        "Module": None,
                                    },
                                    {
                                        "Name": "bind-libs",
                                        "Version": "30:9.3.3-8.el5",
                                        "VersionFormat": "rpm",
                                        "NamespaceName": "ol:5",
                                        "Module": None,
                                    },
                                ],
                            ),
                            "Link": "http://linux.oracle.com/errata/ELSA-2007-0057.html",
                            "Description": """[30:9.3.3-8]
 - added fix for #224445 - CVE-2007-0493 BIND might crash after
   attempting to read free()-ed memory
 - added fix for #225229 - CVE-2007-0494 BIND dnssec denial of service
 - Resolves: rhbz#224445
 - Resolves: rhbz#225229""",
                            "Metadata": {
                                "Issued": "2007-06-26",
                                "RefId": "ELSA-2007-0057",
                                "CVE": [
                                    {
                                        "Name": "CVE-2007-0493",
                                        "Link": "http://linux.oracle.com/cve/CVE-2007-0493.html",
                                    },
                                    {
                                        "Name": "CVE-2007-0494",
                                        "Link": "http://linux.oracle.com/cve/CVE-2007-0494.html",
                                    },
                                ],
                            },
                            "Name": "ELSA-2007-0057",
                            "CVSS": [],
                        },
                    },
                ),
                ("ELSA-2018-4250", "ol:6"): (
                    "501",
                    {
                        "Vulnerability": {
                            "Severity": "High",
                            "NamespaceName": "ol:6",
                            "FixedIn": unordered(
                                [
                                    {
                                        "Name": "kernel-uek-doc",
                                        "Version": "0:2.6.39-400.302.2.el6uek",
                                        "VersionFormat": "rpm",
                                        "NamespaceName": "ol:6",
                                        "Module": None,
                                    },
                                    {
                                        "Name": "kernel-uek",
                                        "Version": "0:2.6.39-400.302.2.el6uek",
                                        "VersionFormat": "rpm",
                                        "NamespaceName": "ol:6",
                                        "Module": None,
                                    },
                                    {
                                        "Name": "kernel-uek-firmware",
                                        "Version": "0:2.6.39-400.302.2.el6uek",
                                        "VersionFormat": "rpm",
                                        "NamespaceName": "ol:6",
                                        "Module": None,
                                    },
                                    {
                                        "Name": "kernel-uek-debug-devel",
                                        "Version": "0:2.6.39-400.302.2.el6uek",
                                        "VersionFormat": "rpm",
                                        "NamespaceName": "ol:6",
                                        "Module": None,
                                    },
                                    {
                                        "Name": "kernel-uek-devel",
                                        "Version": "0:2.6.39-400.302.2.el6uek",
                                        "VersionFormat": "rpm",
                                        "NamespaceName": "ol:6",
                                        "Module": None,
                                    },
                                    {
                                        "Name": "kernel-uek-debug",
                                        "Version": "0:2.6.39-400.302.2.el6uek",
                                        "VersionFormat": "rpm",
                                        "NamespaceName": "ol:6",
                                        "Module": None,
                                    },
                                ],
                            ),
                            "Link": "http://linux.oracle.com/errata/ELSA-2018-4250.html",
                            "Description": """[2.6.39-400.302.2]
- Revert 'Fix up non-directory creation in SGID directories' (Brian Maly)  [Orabug: 28781234]

[2.6.39-400.302.1]
- Fix up non-directory creation in SGID directories (Linus Torvalds)  [Orabug: 28459479]  {CVE-2018-13405}
- ALSA: seq: Make ioctls race-free (Takashi Iwai)  [Orabug: 28459730]  {CVE-2018-7566}
- rds: CVE-2018-7492: Fix NULL pointer dereference in __rds_rdma_map (Hakon Bugge)  [Orabug: 28539910]  {CVE-2018-7492}
- cdrom: Fix info leak/OOB read in cdrom_ioctl_drive_status (Scott Bauer)  [Orabug: 28664549]  {CVE-2018-16658}
- ACPICA: acpi: acpica: fix acpi operand cache leak in nseval.c (Seunghun Han)  [Orabug: 28664580]  {CVE-2017-13695}
- exec: Limit arg stack to at most 75% of _STK_LIM (Kees Cook)  [Orabug: 28710024]  {CVE-2018-14634}""",
                            "Metadata": {
                                "Issued": "2018-10-12",
                                "RefId": "ELSA-2018-4250",
                                "CVE": [
                                    {
                                        "Name": "CVE-2018-7566",
                                        "Link": "http://linux.oracle.com/cve/CVE-2018-7566.html",
                                    },
                                    {
                                        "Name": "CVE-2017-13695",
                                        "Link": "http://linux.oracle.com/cve/CVE-2017-13695.html",
                                    },
                                    {
                                        "Name": "CVE-2018-16658",
                                        "Link": "http://linux.oracle.com/cve/CVE-2018-16658.html",
                                    },
                                    {
                                        "Name": "CVE-2018-7492",
                                        "Link": "http://linux.oracle.com/cve/CVE-2018-7492.html",
                                    },
                                ],
                            },
                            "Name": "ELSA-2018-4250",
                            "CVSS": [],
                        },
                    },
                ),
            },
        ),
    ],
)
def test_parse(tmpdir, helpers, input_file, expected):
    subject = parser.Parser(workspace=workspace.Workspace(tmpdir, "test", create=True))

    mock_data_path = helpers.local_dir(input_file)

    vuln_dict = subject._parse_oval_data(mock_data_path, subject.config)

    assert vuln_dict == expected


class TestKspliceFilterer:
    @pytest.mark.parametrize(
        ("input_vulnerability", "expected_output"),
        [
            (
                {
                    ("ELSA-123", "ol:8"): (
                        "1",
                        {
                            "Vulnerability": {
                                "Name": "ELSA-123",
                                "FixedIn": [
                                    {
                                        "Name": "openssl",
                                        "VersionFormat": "rpm",
                                        "NamespaceName": "ol:8",
                                        "Version": "2:1.1.1g-15.ksplice1.el8_3",
                                        "Module": None,
                                    },
                                ],
                            },
                        },
                    ),
                },
                {
                    ("ELSA-123", "ol:8"): (
                        "1",
                        {
                            "Vulnerability": {
                                "Name": "ELSA-123",
                                "FixedIn": [],
                            },
                        },
                    ),
                },
            ),
            (
                {
                    ("ELSA-123", "ol:8"): (
                        "1",
                        {
                            "Vulnerability": {
                                "Name": "ELSA-123",
                                "FixedIn": [
                                    {
                                        "Name": "openssl",
                                        "VersionFormat": "rpm",
                                        "NamespaceName": "ol:8",
                                        "Version": "2:1.1.1g-15.1.el8_3",
                                        "Module": None,
                                    },
                                ],
                            },
                        },
                    ),
                },
                {
                    ("ELSA-123", "ol:8"): (
                        "1",
                        {
                            "Vulnerability": {
                                "Name": "ELSA-123",
                                "FixedIn": [
                                    {
                                        "Name": "openssl",
                                        "VersionFormat": "rpm",
                                        "NamespaceName": "ol:8",
                                        "Version": "2:1.1.1g-15.1.el8_3",
                                        "Module": None,
                                    },
                                ],
                            },
                        },
                    ),
                },
            ),
            (
                {
                    ("ELSA-123", "ol:8"): (
                        "1",
                        {
                            "Vulnerability": {
                                "Name": "ELSA-123",
                                "FixedIn": [
                                    {
                                        "Name": "openssl",
                                        "VersionFormat": "rpm",
                                        "NamespaceName": "ol:8",
                                        "Version": "2:1.1.1g-15.ksplice1.el8_3",
                                        "Module": None,
                                    },
                                    {
                                        "Name": "openssl",
                                        "VersionFormat": "rpm",
                                        "NamespaceName": "ol:8",
                                        "Version": "2:1.1.1g-15.1.el8_3",
                                        "Module": None,
                                    },
                                ],
                            },
                        },
                    ),
                },
                {
                    ("ELSA-123", "ol:8"): (
                        "1",
                        {
                            "Vulnerability": {
                                "Name": "ELSA-123",
                                "FixedIn": [
                                    {
                                        "Name": "openssl",
                                        "VersionFormat": "rpm",
                                        "NamespaceName": "ol:8",
                                        "Version": "2:1.1.1g-15.1.el8_3",
                                        "Module": None,
                                    },
                                ],
                            },
                        },
                    ),
                },
            ),
        ],
    )
    def test_filterer(self, input_vulnerability, expected_output):
        f = parser.KspliceFilterer()
        assert f.filter(input_vulnerability) == expected_output


def test_provider_schema(helpers, disable_get_requests, monkeypatch):
    workspace = helpers.provider_workspace_helper(name=Provider.name())

    c = Config()
    c.runtime.result_store = result.StoreStrategy.FLAT_FILE
    p = Provider(root=workspace.root, config=c)

    mock_data_path = helpers.local_dir("test-fixtures/mock_data")
    shutil.copy(mock_data_path, workspace.input_dir / "com.oracle.elsa-all.xml")

    def mock_download():
        return None

    monkeypatch.setattr(p.parser, "_download", mock_download)

    p.update(None)

    assert workspace.num_result_entries() == 2
    assert workspace.result_schemas_valid(require_entries=True)


def test_provider_via_snapshot(helpers, disable_get_requests, monkeypatch):
    workspace = helpers.provider_workspace_helper(name=Provider.name())

    c = Config()
    # keep all of the default values for the result store, but override the strategy
    c.runtime.result_store = result.StoreStrategy.FLAT_FILE
    p = Provider(
        root=workspace.root,
        config=c,
    )

    mock_data_path = helpers.local_dir("test-fixtures/mock_data")
    shutil.copy(mock_data_path, workspace.input_dir / "com.oracle.elsa-all.xml")

    def mock_download():
        return None

    monkeypatch.setattr(p.parser, "_download", mock_download)

    p.update(None)

    workspace.assert_result_snapshots()



from __future__ import annotations

import logging
import os

import orjson
import requests

from vunnel import utils, workspace

# NOTE, CHANGE ME!: this namespace should be unique to your provider and match expectations from
# grype to know what to search for in the DB.
NAMESPACE = "GRYPEOSNAMESPACETHATYOUCHOOSE"


class Parser:
    # NOTE, CHANGE ME!: remove / add / change these attributes as needed to download and parse your provider
    _json_url_ = "https://services.nvd.nist.gov/made-up-location"
    _json_file_ = "vulnerability_data.json"

    def __init__(self, ws: workspace.Workspace, download_timeout: int = 125, logger: logging.Logger | None = None):
        self.workspace = ws
        self.download_timeout = download_timeout
        self.json_file_path = os.path.join(ws.input_path, self._json_file_)

        # NOTE, CHANGE ME!: you should always record any URLs accessed in this list, either
        # statically or dynamically within _download()
        self.urls = [self._json_url_]

        if not logger:
            logger = logging.getLogger(self.__class__.__name__)
        self.logger = logger

    def get(self):
        self._download()
        yield from self._normalize()

    @utils.retry_with_backoff()
    def _download(self):
        self.logger.info(f"downloading vulnerability data from {self._json_url_}")

        r = requests.get(self._json_url_, timeout=self.download_timeout)
        r.raise_for_status()

        with open(self.json_file_path, "w", encoding="utf-8") as f:
            f.write(r.text)

    def _normalize(self):

        with open(self.json_file_path, encoding="utf-8") as f:

            for input_record in orjson.loads(f.read()):

                vuln_id = input_record["name"]

                # NOTE: this is in the data shape described by the OS vulnerability schema
                yield vuln_id, {
                    "Vulnerability": {
                        "Name": vuln_id,
                        "NamespaceName": NAMESPACE,
                        "Link": f"https://someplace.com/{vuln_id}",
                        "Severity": input_record["severity"],
                        "Description": input_record["description"],
                        "FixedIn": [
                            {
                                "Name": p,
                                "VersionFormat": "apk",
                                "NamespaceName": NAMESPACE,
                                "Version": input_record["fixed"] or "None",
                            }
                            for p in input_record["packages"]
                        ],
                    },
                }



from __future__ import annotations

import copy
import glob
import logging
import os
import re
from collections import namedtuple
from typing import Any

import orjson

from vunnel.result import SQLiteReader
from vunnel.utils import http_wrapper as http
from vunnel.utils import vulnerability

DSAFixedInTuple = namedtuple("DSAFixedInTuple", ["dsa", "link", "distro", "pkg", "ver"])
DSACollection = namedtuple("DSACollection", ["cves", "nocves"])


# Only releases presenting this mapping will be output by the driver, maintain it with new releases.
# Can also be extended via configuration.
debian_distro_map = {
    "trixie": "13",
    "bookworm": "12",
    "bullseye": "11",
    "buster": "10",
    "stretch": "9",
    "jessie": "8",
    "wheezy": "7",
    "sid": "unstable",
}


class Parser:
    _json_url_ = "https://security-tracker.debian.org/tracker/data/json"
    _dsa_url_ = "https://salsa.debian.org/security-tracker-team/security-tracker/raw/master/data/DSA/list"
    _json_file_ = "debian.json"
    _dsa_file_ = "DSA"

    _dsa_start_regex_ = re.compile(r"^\S+.*")
    _dsa_info_regex_ = re.compile(r"^\[(.*)\]\s+(DSA-[^\s]+)\s+([^\s]+)[-\s]+(.*)")
    _cve_list_regex_ = re.compile(r"^\s+\{(.*)\}")
    _fixed_in_regex_ = re.compile(r"^\s+\[(.*)\][-\s]+([^\s]*)(.*)")
    _fixed_in_note_regex_ = re.compile(r"^\s+NOTE:\s+\[(.*)\][-\s]+([^\s]*)(.*)")
    _base_dsa_id_regex_ = re.compile(r"(DSA-[^-]+).*")

    def __init__(self, workspace, download_timeout=125, logger=None, distro_map=None):
        self.workspace = workspace
        self.download_timeout = download_timeout
        if not distro_map:
            distro_map = debian_distro_map
        self.debian_distro_map = distro_map
        self.json_file_path = os.path.join(workspace.input_path, self._json_file_)
        self.dsa_file_path = os.path.join(workspace.input_path, self._dsa_file_)
        self.legacy_records_path = os.path.join(self.workspace.input_path, "legacy")
        self.urls = [self._json_url_, self._dsa_url_]

        if not logger:
            logger = logging.getLogger(self.__class__.__name__)
        self.logger = logger

    def _download_json(self):
        """
        Downloads debian json file
        :return:
        """
        try:
            r = http.get(self._json_url_, self.logger, timeout=self.download_timeout)
            orjson.loads(r.text)  # quick check if json is valid
            with open(self.json_file_path, "w", encoding="utf-8") as OFH:
                OFH.write(r.text)

        except Exception:
            self.logger.exception("Error downloading debian json file")
            raise

    def _download_dsa(self):
        """
        Downloads debian dsa file
        :return:
        """
        try:
            r = http.get(self._dsa_url_, self.logger, timeout=self.download_timeout)
            with open(self.dsa_file_path, "w", encoding="utf-8") as OFH:
                OFH.write(r.text)

        except Exception:
            self.logger.exception("error downloading debian DSA file")
            raise

    def _get_cve_to_dsalist(self, dsa):
        """

        :param dsa:
        :return:
        """
        ns_cve_dsalist = {}
        if dsa["cves"]:
            for fixedin in dsa["fixed_in"]:
                for cve in dsa["cves"]:
                    ns = fixedin["distro"]
                    if ns not in ns_cve_dsalist:
                        ns_cve_dsalist[ns] = {cve: []}
                    if cve not in ns_cve_dsalist[ns]:
                        ns_cve_dsalist[ns][cve] = []
                    ns_cve_dsalist[ns][cve].append(
                        DSAFixedInTuple(
                            dsa=dsa["id"],
                            link=dsa["link"],
                            distro=fixedin["distro"],
                            pkg=fixedin["pkg"],
                            ver=fixedin["ver"],
                        ),
                    )
        else:
            self.logger.debug(f"no CVEs found for {dsa['id']}")

        return ns_cve_dsalist

    def _parse_dsa_record(self, dsa_lines):  # noqa: C901
        """

        :param dsa_lines:
        :return:
        """
        try:
            dsa = {
                "cves": [],
                "fixed_in": [],
            }
            dsa_info_matched = False
            cve_list_matched = False

            for line in dsa_lines:
                if not dsa_info_matched:
                    dsa_info_match = re.match(self._dsa_info_regex_, line)
                    if dsa_info_match:
                        date, vuln_id, package, summary = dsa_info_match.groups()
                        summary = summary.strip() if summary else None
                        dsa["date"] = date
                        dsa["id"] = vuln_id
                        dsa["package"] = package
                        dsa["summary"] = summary
                        dsa["link"] = f"https://security-tracker.debian.org/tracker/{vuln_id}"
                        dsa_info_matched = True
                        continue

                if not cve_list_matched:
                    cve_list_match = re.match(self._cve_list_regex_, line)
                    if cve_list_match:
                        cves = cve_list_match.group(1).split()
                        dsa["cves"] = cves
                        cve_list_matched = True
                        continue

                fixed_in_match = re.match(self._fixed_in_regex_, line)
                if fixed_in_match:
                    distro, pkg, version = fixed_in_match.groups()
                    version = version.strip() if version else None
                    if not version:
                        self.logger.debug(
                            f"release version not included dsa: {dsa.get('id')}, distro: {distro}, pkg: {pkg}",
                        )
                    dsa["fixed_in"].append({"distro": distro, "pkg": pkg, "ver": version})
                    continue

                fixed_in_note_match = re.match(self._fixed_in_note_regex_, line)
                if fixed_in_note_match:
                    distro, pkg, version = fixed_in_note_match.groups()
                    version = version.strip() if version else None
                    if not version:
                        self.logger.debug(
                            f"release version not included dsa: {dsa.get('id')}, distro: {distro}, pkg: {pkg}",
                        )
                    dsa["fixed_in"].append({"distro": distro, "pkg": pkg, "ver": version})
                    continue

            return dsa
        except Exception:
            self.logger.exception("failed to parse dsa record")

    def _get_dsa_map(self):
        """

        :return:
        """

        if os.path.exists(self.dsa_file_path):
            dsa_map = {}

            with open(self.dsa_file_path, encoding="utf-8") as fp:
                dsa_rec = []
                line = fp.readline()
                while line:
                    if re.match(self._dsa_start_regex_, line) and dsa_rec:
                        # process bunch
                        dsa = self._parse_dsa_record(dsa_lines=dsa_rec)

                        # gather all DSAs with the same base DSA ID so the missing CVEs on some DSAs can be filled in
                        base_dsa_id_match = re.match(self._base_dsa_id_regex_, dsa["id"])
                        base_dsa_id = base_dsa_id_match.group(1) if base_dsa_id_match else dsa["id"]

                        if base_dsa_id not in dsa_map:
                            dsa_map[base_dsa_id] = DSACollection(cves=[], nocves=[])

                        if dsa["cves"]:
                            dsa_map[base_dsa_id].cves.append(dsa)
                        else:
                            dsa_map[base_dsa_id].nocves.append(dsa)
                        del dsa_rec[:]

                    dsa_rec.append(line)
                    line = fp.readline()

            return dsa_map

        raise Exception(f"debian DSA source not found under {self.dsa_file_path}")

    def _normalize_dsa_list(self):
        ns_cve_dsalist = {}
        dsa_map = self._get_dsa_map()

        for dsa_collection in dsa_map.values():
            dsas_with_cves = dsa_collection.cves

            # Fill in the missing CVEs for some incremental DSAs if you can.
            # This also filters out DSAs that don't have neither CVEs nor fixed_in records
            if dsa_collection.nocves and dsas_with_cves:
                # get the cves from any dsa in the list
                cve_list = dsas_with_cves[0]["cves"]
                for dsa in dsa_collection.nocves:
                    dsa.update({"cves": copy.deepcopy(cve_list)})

                dsas_with_cves.extend(dsa_collection.nocves)

            # Invert the data and map DSAs to CVEs to namespaces
            for dsa in dsas_with_cves:
                for ns, cve_dsalist in self._get_cve_to_dsalist(dsa).items():
                    if ns not in ns_cve_dsalist:
                        ns_cve_dsalist[ns] = {}
                    for cve, dsalist in cve_dsalist.items():
                        if cve not in ns_cve_dsalist[ns]:
                            ns_cve_dsalist[ns][cve] = []
                        ns_cve_dsalist[ns][cve].extend(dsalist)

        dsa_map.clear()

        return ns_cve_dsalist

    def _normalize_json(self, ns_cve_dsalist=None):  # noqa: PLR0912,PLR0915,C901
        adv_mets = {}
        # all_matched_dsas = set()
        # all_dsas = set()

        # normalize DSA list
        # ns_cve_dsalist = normalize_dsa_list(dsaloc)

        # # metrics colleciton
        # for ns, cve_dsalist in ns_cve_dsalist.iteritems():
        #     if common.debian_distro_map.get(ns, None):
        #         all_dsas |= set([dsa_tup.dsa for dsalist in cve_dsalist.values() for dsa_tup in dsalist])

        if os.path.exists(self.json_file_path):
            with open(self.json_file_path, encoding="utf-8") as FH:
                data = orjson.loads(FH.read())
        else:
            raise Exception(f"debian json source not found under {self.json_file_path}")

        if ns_cve_dsalist is None:
            ns_cve_dsalist = {}

        vuln_records = {}

        for pkg in data:
            for vid in data[pkg]:
                # skip non CVE vids
                if not re.match("^CVE.*", vid):
                    continue

                # gather NVD data for this CVE. Pulling this logic out of the for loop as NVD data remains the same
                # regardless of the debian release.
                nvd_severity = None
                # if session:
                #     try:
                #         nvd_severity = nvd.get_severity(vid, session=session)
                #     except Exception:
                #         self.logger.exception("ignoring error processing nvdv2 record")

                vulnerability_data = data[pkg][vid]
                release_data = vulnerability_data.get("releases")
                if not isinstance(release_data, dict) or not release_data:
                    continue

                for rel, distro_record in release_data.items():
                    try:
                        relno = self.debian_distro_map.get(rel)
                        if not relno:
                            continue

                        # if relno not in retlists:
                        #    retlists[relno] = []

                        if relno not in vuln_records:
                            vuln_records[relno] = {}

                        complete = True
                        if distro_record.get("status", "") == "undetermined":
                            complete = False

                        if complete:
                            if vid not in vuln_records[relno]:
                                # create a new record
                                vuln_records[relno][vid] = copy.deepcopy(vulnerability.vulnerability_element)
                                vuln_record = vuln_records[relno][vid]

                                # populate the static information about the new vuln record
                                vuln_record["Vulnerability"]["Description"] = vulnerability_data.get("description", "")
                                vuln_record["Vulnerability"]["Name"] = str(vid)
                                vuln_record["Vulnerability"]["NamespaceName"] = "debian:" + str(relno)
                                vuln_record["Vulnerability"]["Link"] = "https://security-tracker.debian.org/tracker/" + str(vid)
                                vuln_record["Vulnerability"]["Severity"] = "Unknown"
                            else:
                                vuln_record = vuln_records[relno][vid]

                            # set severity
                            # from https://anonscm.debian.org/viewvc/secure-testing/bin/tracker_service.py
                            sev = None
                            if "urgency" in distro_record:
                                if distro_record["urgency"] in ["low", "low**"]:
                                    sev = "Low"
                                elif distro_record["urgency"] in ["medium", "medium**"]:
                                    sev = "Medium"
                                elif distro_record["urgency"] in ["high", "high**"]:
                                    sev = "High"
                                elif distro_record["urgency"] in [
                                    "unimportant",
                                    "end-of-life",
                                ]:
                                    sev = "Negligible"
                                elif nvd_severity:  # no match to urgency found
                                    sev = nvd_severity  # fallback to nvd severity
                                else:
                                    sev = "Unknown"
                            elif nvd_severity:  # urgency element is not present
                                sev = nvd_severity  # fallback to nvd severity
                            else:
                                sev = "Unknown"

                            if sev and vulnerability.severity_order[sev] > vulnerability.severity_order[vuln_record["Vulnerability"]["Severity"]]:
                                vuln_record["Vulnerability"]["Severity"] = sev

                            # add fixedIn
                            skip_fixedin = False
                            fixed_el = {
                                "Name": pkg,
                                "NamespaceName": "debian:" + str(relno),
                                "VersionFormat": "dpkg",
                            }

                            if "fixed_version" in distro_record:
                                fixed_el["Version"] = distro_record["fixed_version"]
                                if distro_record["fixed_version"] == "0":
                                    # version == 0 should mean that the
                                    # package was determined to not be
                                    # vulnerable in the distro namespace
                                    # (from reviewing
                                    # https://security-tracker.debian.org/tracker/)
                                    skip_fixedin = True
                            else:
                                fixed_el["Version"] = "None"

                            if not skip_fixedin:
                                # collect metrics for vendor advisory
                                met_ns = vuln_record["Vulnerability"]["NamespaceName"]
                                met_sev = vuln_record["Vulnerability"]["Severity"]

                                if met_ns not in adv_mets:
                                    adv_mets[met_ns] = {
                                        met_sev: {
                                            "dsa": {"fixed": 0, "notfixed": 0},
                                            "nodsa": {"fixed": 0, "notfixed": 0},
                                            "neither": {"fixed": 0, "notfixed": 0},
                                        },
                                    }

                                if met_sev not in adv_mets[met_ns]:
                                    adv_mets[met_ns][met_sev] = {
                                        "dsa": {"fixed": 0, "notfixed": 0},
                                        "nodsa": {"fixed": 0, "notfixed": 0},
                                        "neither": {"fixed": 0, "notfixed": 0},
                                    }

                                # find DSAs associated with the CVE and package in the namespace
                                matched_dsas = [dsa for dsa in ns_cve_dsalist.get(rel, {}).get(vid, []) if dsa.pkg == pkg]

                                # add vendor advisory information to the fixed in record
                                if matched_dsas:
                                    fixed_el["VendorAdvisory"] = {
                                        "NoAdvisory": False,
                                        "AdvisorySummary": [{"ID": x.dsa, "Link": x.link} for x in matched_dsas],
                                    }
                                    # all_matched_dsas |= set([x.dsa for x in matched_dsas])
                                    adv_mets[met_ns][met_sev]["dsa"]["notfixed" if fixed_el["Version"] == "None" else "fixed"] += 1
                                elif "nodsa" in distro_record:
                                    fixed_el["VendorAdvisory"] = {"NoAdvisory": True}
                                    adv_mets[met_ns][met_sev]["nodsa"]["notfixed" if fixed_el["Version"] == "None" else "fixed"] += 1
                                else:
                                    fixed_el["VendorAdvisory"] = {
                                        "NoAdvisory": False,
                                        "AdvisorySummary": [],
                                    }
                                    adv_mets[met_ns][met_sev]["neither"]["notfixed" if fixed_el["Version"] == "None" else "fixed"] += 1

                                # append fixed in record to vulnerability
                                vuln_record["Vulnerability"]["FixedIn"].append(fixed_el)

                            # strip out any top level that is not set
                            final_record = {"Vulnerability": {}}
                            for k in vuln_record["Vulnerability"]:
                                if vuln_record["Vulnerability"][k]:
                                    final_record["Vulnerability"][k] = copy.deepcopy(vuln_record["Vulnerability"][k])

                            # retlists[relno].append(final_record)

                    except Exception:
                        self.logger.exception(f"ignoring error parsing vuln: {vid}, pkg: {pkg}, rel: {rel}")

        self.logger.debug(f"metrics for advisory information: {orjson.dumps(adv_mets).decode('utf-8')}")

        adv_mets.clear()
        # all_dsas.clear()
        # all_matched_dsas.clear()

        return vuln_records

    def _get_legacy_records(self) -> dict[str, dict[str, Any]]:
        legacy_records = self._get_legacy_records_from_results_db()

        fs_legacy_records = self._get_legacy_records_from_feed_service_datadrop()
        for relno, vuln_dict in fs_legacy_records.items():
            if relno not in legacy_records:
                legacy_records[relno] = {}
            legacy_records[relno].update(vuln_dict)

        if legacy_records:
            self.logger.info(f"found existing legacy data for the following releases: {list(legacy_records.keys())}")
        else:
            self.logger.info("no existing legacy data found")

        return legacy_records

    def _get_legacy_records_from_results_db(self) -> dict[str, dict[str, Any]]:
        legacy_records = {}

        def process_result(file_path: str) -> None:
            self.logger.info(f"found existing legacy dataset: {file_path}")

            releases = set()
            records = 0
            with SQLiteReader(file_path) as db:
                envelopes = db.read_all()
                for envelope in envelopes:
                    relno = envelope.item["Vulnerability"]["NamespaceName"].split(":")[-1]
                    releases.add(relno)
                    vid = envelope.item["Vulnerability"]["Name"]
                    if relno not in legacy_records:
                        legacy_records[relno] = {}

                    records += 1
                    legacy_records[relno][vid] = envelope.item

            self.logger.debug(f"legacy dataset {file_path} contains {len(releases)} releases with {records} records")

        result_files = glob.glob(os.path.join(self.legacy_records_path, "**", "results.db"), recursive=True)

        for file_path in result_files:
            process_result(file_path)

        return legacy_records

    def _get_legacy_records_from_feed_service_datadrop(self) -> dict[str, dict[str, Any]]:
        legacy_records = {}

        def process_file(contents: list[dict[str, Any]]) -> None:
            for record in contents:
                relno = record["Vulnerability"]["NamespaceName"].split(":")[-1]
                vid = record["Vulnerability"]["Name"]
                if relno not in legacy_records:
                    legacy_records[relno] = {}

                # ensure results are compliant with the current schema
                cvss_metadata = record["Vulnerability"].get("Metadata", {}).get("NVD", {}).get("CVSSv2", {})
                if cvss_metadata:
                    if cvss_metadata["Vectors"] is None:
                        del cvss_metadata["Vectors"]
                    record["Vulnerability"]["Metadata"]["NVD"]["CVSSv2"] = cvss_metadata

                # write the record back
                legacy_records[relno][vid] = record

        # read every json file in the legacy directory
        for root, _dirs, files in os.walk(self.legacy_records_path):
            for file in files:
                if file.endswith(".json") and file.startswith("vulnerabilities"):
                    with open(os.path.join(root, file)) as f:
                        process_file(orjson.loads(f.read()))

        if legacy_records:
            self.logger.info(f"found feed service legacy data for the following releases: {list(legacy_records.keys())}")

        return legacy_records

    def get(self):
        # download the files
        self._download_json()
        self._download_dsa()

        # normalize dsa list first
        ns_cve_dsalist = self._normalize_dsa_list()

        # normalize json file
        vuln_records = self._normalize_json(ns_cve_dsalist=ns_cve_dsalist)

        # fetch records from legacy (if they exist)
        legacy_records = self._get_legacy_records()
        for relno, vuln_dict in legacy_records.items():
            if relno not in vuln_records:
                vuln_records[relno] = {}
            vuln_records[relno].update(vuln_dict)

        if vuln_records:
            for relno, vuln_dict in vuln_records.items():
                for vid, vuln_record in vuln_dict.items():
                    # HACK: when we can represent per-package severity or have a good mechanism
                    # for overriding upstream data, we should take this out.
                    severity_override = {"CVE-2020-35525", "CVE-2020-35527", "CVE-2023-4863", "CVE-2023-44487"}
                    if vid in severity_override:
                        self.logger.info(
                            f"clearing severity on {vid}, see https://github.com/anchore/grype-db/issues/108#issuecomment-1796301073",
                        )
                        vuln_record["Vulnerability"]["Severity"] = "Unknown"
                    yield relno, vid, vuln_record
        else:
            yield from ()



import json
import os

import pytest

from vunnel import result, workspace
from vunnel.providers import epss
from vunnel.providers.epss import manager


@pytest.mark.parametrize(
    ("mock_data_path", "expected_vulns"),
    [
        ("test-fixtures/example-1.csv",
            [
                ('CVE-1999-0001',{'cve': 'CVE-1999-0001','date': '2025-02-20','epss': 0.00383,'percentile': 0.73075}),
                ('CVE-1999-0002',{'cve': 'CVE-1999-0002','date': '2025-02-20','epss': 0.01328,'percentile': 0.85996}),
                ('CVE-1999-0003',{'cve': 'CVE-1999-0003','date': '2025-02-20','epss': 0.04409,'percentile': 0.9244}),
                ('CVE-1999-0004',{'cve': 'CVE-1999-0004','date': '2025-02-20','epss': 0.00917,'percentile': 0.82963}),
                ('CVE-1999-0005',{'cve': 'CVE-1999-0005','date': '2025-02-20','epss': 0.91963,'percentile': 0.99214}),
            ],
        ),
    ],
)
def test_parser(tmpdir, helpers, mocker,  mock_data_path, expected_vulns):
    mock_data_path = helpers.local_dir(mock_data_path)

    subject = manager.Manager(
        dataset="current",
        ws=workspace.Workspace(tmpdir, "test", create=True),
        url_template="http://localhost/nowhere/epss_scores-{}.csv.gz",
    )

    subject._download = mocker.Mock(return_value=mock_data_path)

    actual_vulns = list(subject.get())

    assert expected_vulns == actual_vulns


@pytest.mark.parametrize(
    ("mock_data_path", "expected_written_entries"),
    [
        ("test-fixtures/example-1.csv", 5),
    ],
)
def test_provider_schema(helpers, mock_data_path, expected_written_entries, disable_get_requests, mocker):
    workspace = helpers.provider_workspace_helper(name=epss.Provider.name())
    mock_data_path = helpers.local_dir(mock_data_path)

    # Setup provider
    c = epss.Config()
    p = epss.Provider(root=workspace.root, config=c)
    p.config.runtime.result_store = result.StoreStrategy.FLAT_FILE
    p.manager._download = mocker.Mock(return_value=mock_data_path)

    p.update(None)

    assert expected_written_entries == workspace.num_result_entries()
    assert workspace.result_schemas_valid(require_entries=expected_written_entries > 0)


@pytest.mark.parametrize(
    ("mock_data_path",),
    [
        ("test-fixtures/example-1.csv",),
    ],
)
def test_provider_via_snapshot(helpers, mock_data_path, disable_get_requests, mocker):
    snapshot_name = os.path.basename(mock_data_path).split(".")[0]

    workspace = helpers.provider_workspace_helper(
        name=epss.Provider.name(),
        snapshot_prefix=snapshot_name,
    )
    mock_data_path = helpers.local_dir(mock_data_path)

    c = epss.Config()
    p = epss.Provider(
        root=workspace.root,
        config=c,
    )
    p.config.runtime.result_store = result.StoreStrategy.FLAT_FILE
    p.manager._download = mocker.Mock(return_value=mock_data_path)

    p.update(None)

    workspace.assert_result_snapshots()


@pytest.mark.parametrize(
    ("test_file", "exception_type"),
    [
        pytest.param(
            "test-fixtures/missing-date.csv",
            ValueError,
            id="missing-date-header",
        ),
        pytest.param(
            "test-fixtures/invalid-header.csv",
            ValueError,
            id="invalid-csv-header",
        ),
    ],
)
def test_manager_normalize_invalid_data(
    tmpdir, helpers, test_file, exception_type
):
    ws = workspace.Workspace(tmpdir, "test", create=True)
    subject = manager.Manager(dataset="current", ws=ws)

    test_file_path = helpers.local_dir(test_file)

    with pytest.raises(exception_type):
        list(subject._normalize(test_file_path))


def test_count_lines(tmpdir):
    test_file = os.path.join(tmpdir, "test.txt")
    content = "line1\nline2\nline3"

    with open(test_file, "w") as f:
        f.write(content)

    assert manager._count_lines(test_file) == 3



from __future__ import annotations

import json
import os
from pathlib import Path
import subprocess
import uuid
import os.path
import shutil

import jsonschema
import orjson
import pytest


class WorkspaceHelper:
    def __init__(self, root: str, name: str, snapshot):
        self.root = root
        self.name = name
        self.snapshot = snapshot

    @property
    def metadata_path(self):
        return self.root / self.name / "metadata.json"

    @property
    def input_dir(self):
        return self.root / self.name / "input"

    @property
    def input_path(self) -> str:
        return self.root / self.name / "input"

    @property
    def results_dir(self):
        return self.root / self.name / "results"

    def result_files(self):
        results = []
        for root, _dirs, files in os.walk(self.results_dir):
            for filename in files:
                results.append(os.path.join(root, filename))
        return results

    def _snapshot_files(self):
        snapshot_files = []

        for root, _dirs, files in os.walk(self.snapshot.snapshot_dir):
            for filename in files:
                snapshot_files.append(os.path.join(root, filename))
        return snapshot_files

    def num_result_entries(self):
        return len(self.result_files())

    def result_schemas_valid(self, require_entries: bool = True) -> bool:
        entries_validated = 0
        for result_file in self.result_files():
            with open(result_file) as f:
                envelope = json.load(f)
                schema_url = envelope["schema"]

                schema_dict = load_json_schema(get_schema_repo_path(schema_url))
                _validate_json_schema(instance=envelope["item"], schema=schema_dict)
                entries_validated += 1

        if require_entries and entries_validated == 0:
            raise ValueError("no entries were validated")

        return True

    def metadata_schema_valid(self) -> bool:
        with open(self.metadata_path) as f:
            item = json.load(f)
            schema_url = item["schema"]["url"]

            schema_dict = load_json_schema(get_schema_repo_path(schema_url))
            _validate_json_schema(instance=item, schema=schema_dict)

        return True

    def copy_input_fixtures(self, mock_data_path: str):
        shutil.copytree(mock_data_path, self.input_dir, dirs_exist_ok=True)

    def assert_result_snapshots(self):
        expected_files_to_test = set(self._snapshot_files())
        missing_snapshot_files = []

        for result_file in self.result_files():
            # protection against test configuration not swapping to the flat file store strategy
            assert result_file.endswith(".json")

            with open(result_file) as f:
                snapshot_path = result_file.split("results/")[-1]

                snapshot_abs_path = os.path.join(self.snapshot.snapshot_dir, snapshot_path)

                if not self.snapshot._snapshot_update and not os.path.exists(snapshot_abs_path):
                    missing_snapshot_files.append(snapshot_abs_path)
                else:
                    d = orjson.loads(f.read())
                    expected_bytes = orjson.dumps(
                        d, option=orjson.OPT_APPEND_NEWLINE | orjson.OPT_INDENT_2 | orjson.OPT_SORT_KEYS
                    )
                    self.snapshot.assert_match(expected_bytes, snapshot_path)

                    if snapshot_abs_path in expected_files_to_test:
                        expected_files_to_test.remove(snapshot_abs_path)

        message_lines = []
        if expected_files_to_test:
            message_lines.append("existing snapshot files that were not asserted:")
            for expected_snapshot_path in expected_files_to_test:
                message_lines.append(f"  - {expected_snapshot_path}")

        if missing_snapshot_files:
            if message_lines:
                message_lines.append("")
            message_lines.append("missing snapshot files:")
            for missing_snapshot_file in missing_snapshot_files:
                message_lines.append(f"  - {missing_snapshot_file}")

        if message_lines:
            pytest.fail("\n".join(message_lines), pytrace=False)


@pytest.fixture()
def validate_json_schema():
    def apply(content: str):
        doc = json.loads(content)
        schema_url = doc.get("schema", {}).get("url")
        if not schema_url:
            raise ValueError("No schema URL found in document")

        schema_path = get_schema_repo_path(schema_url)
        schema = load_json_schema(schema_path)
        _validate_json_schema(instance=doc, schema=schema)

    return apply


def load_json_schema(path: str) -> dict:
    with open(path) as f:
        return json.load(f)


def get_schema_repo_path(url: str) -> str:
    # e.g. https://raw.githubusercontent.com/anchore/vunnel/main/schema/vulnerability/nvd/schema-{version}.json
    relative_path = url.removeprefix("https://raw.githubusercontent.com/anchore/vunnel/main/")
    if relative_path == url:
        raise ValueError(f"URL {url!r} is not a valid schema URL")
    return os.path.join(git_root(), relative_path)


class Helpers:
    def __init__(self, request, tmpdir, snapshot):
        # current information about the running test
        # docs: https://docs.pytest.org/en/6.2.x/reference.html#std-fixture-request
        self.request = request
        self.tmpdir = tmpdir
        self.snapshot = snapshot

    def local_dir(self, path: str):
        """
        Returns the path of a file relative to the current test file.

        Given the following setup:

            test/unit/providers/centos/
            ├── test-fixtures
            │   ├── mock_data_1
            │   └── mock_data_2
            └── test_centos.py

        The call `local_dir("test-fixtures/mock_data_1")` will return the absolute path to
        the mock data file relative to test_centos.py
        """
        current_test_filepath = os.path.realpath(self.request.module.__file__)
        parent = os.path.realpath(os.path.dirname(current_test_filepath))
        return os.path.join(parent, path)

    def provider_workspace_helper(
        self, name: str, create: bool = True, input_fixture: str | None = None, snapshot_prefix: str = ""
    ) -> WorkspaceHelper:
        root = self.tmpdir
        if create:
            os.makedirs(root / name / "input")
            os.makedirs(root / name / "results")

        # any snapshot tests should be stored in the same place
        snapshot_path = "test-fixtures/snapshots"
        if snapshot_prefix:
            snapshot_path = os.path.join(snapshot_path, snapshot_prefix)
        self.snapshot.snapshot_dir = self.local_dir(snapshot_path)

        h = WorkspaceHelper(root, name, self.snapshot)

        if input_fixture:
            h.copy_input_fixtures(self.local_dir(input_fixture))

        return h


@pytest.fixture()
def helpers(request, tmpdir, snapshot):
    """
    Returns a common set of helper functions for tests.
    """
    return Helpers(request, tmpdir, snapshot)


def git_root() -> str:
    return (
        subprocess.Popen(["git", "rev-parse", "--show-toplevel"], stdout=subprocess.PIPE)
        .communicate()[0]
        .rstrip()
        .decode("utf-8")
    )


@pytest.fixture()
def dummy_file():
    def apply(d: str, name: str = ""):
        if name == "":
            suffix = str(uuid.uuid4())[:8]
            name = f"random-{suffix}.json"

        path = os.path.join(d, name)
        with open(path, "w", encoding="utf-8") as f:
            json.dump({"name": name}, f)
        return path

    return apply


@pytest.fixture()
def disable_get_requests(monkeypatch):
    def disabled(*args, **kwargs):
        raise RuntimeError("requests disabled but HTTP GET attempted")

    from vunnel import utils

    return monkeypatch.setattr(utils.http_wrapper, "get", disabled)


def _validate_json_schema(instance: dict, schema: dict):
    _schema_validator(schema=schema).validate(instance=instance)


def _schema_validator(schema: dict) -> jsonschema.Draft7Validator:
    from referencing import Registry, Resource

    # load up known schema references into a common registry to prevent network calls
    # see https://python-jsonschema.readthedocs.io/en/latest/referencing/ for more details

    paths = {
        "schema/vulnerability/nvd/cvss/schema-v2.0.json": "https://csrc.nist.gov/schema/nvd/api/2.0/external/cvss-v2.0.json",
        "schema/vulnerability/nvd/cvss/schema-v3.0.json": "https://csrc.nist.gov/schema/nvd/api/2.0/external/cvss-v3.0.json",
        "schema/vulnerability/nvd/cvss/schema-v3.1.json": "https://csrc.nist.gov/schema/nvd/api/2.0/external/cvss-v3.1.json",
    }

    registry = Registry()
    for path, url in paths.items():
        in_repo_path = os.path.join(git_root(), path)
        schema_resource = Resource.from_contents(load_json_schema(in_repo_path))
        registry = registry.with_resource(url, schema_resource)

    return jsonschema.Draft7Validator(schema, registry=registry)



from __future__ import annotations

import concurrent.futures
import copy
import enum
import logging
import os
import re
import time
from dataclasses import asdict, dataclass, field
from typing import TYPE_CHECKING, Any

import orjson

from .git import GitWrapper

if TYPE_CHECKING:
    from collections.abc import Generator

    from vunnel.workspace import Workspace

namespace = "ubuntu"

default_max_workers = 8
default_git_url = "git://git.launchpad.net/ubuntu-cve-tracker"
default_git_branch = "master"

ubuntu_pkg_version_format = "dpkg"
ubuntu_cve_url = "https://ubuntu.com/security/{}"
_patches_header_regex = re.compile(r"Patches_(\S+)\s*")
_patches_regex = re.compile(r"\s*(.+)_(.+)\s*:\s+(.+)\s*")
_patch_state_regex = re.compile(r"\s*(\S+)(\s+.+)?\s*")
_indent_line_regex = re.compile(r"\s+\S+\s*")
_package_priority_regex = re.compile(r"\s*Priority_(\S+)\s*:\s+(\S+)\s*")
_cve_filename_regex = re.compile("CVE-[0-9]+-[0-9]+")

# Per the Ubuntu README in the security tracker BZR repo:
# Maps the state name to whether it indicates a package is vulnerable
patch_states = {
    "DNE": False,  # Does Not Exist, the package is does not exist in a particular ubuntu release
    "needs-triage": True,  # Not yet determined if CVE affects package, consider all versions vulnerable until determination is made
    "ignored": False,  # CVE does not affect the package or no updates (e.g. end-of-life) (NOTE: should still report?)
    "not-affected": False,  # The package is related to the issue, but not affected by it.
    "needed": True,  # Package is vuln and needs a fix. No version yet.
    "released": True,  # The package is affected and a fix has been released with the given version.
    "pending": True,
    # Indicates a vuln with release coming in the expected version. We ignore here for now assuming it will be in 'released' status shortly.
    "active": True,  # The package is affected, needs fixing and is being worked on.
    "deferred": True,
    # The package is affected by work on fix is deferred for some reason. In this case, ignore any version info as it may be random (e.g. a date)
}

patch_merge_criteria = {
    "status": re.compile(r"ignored"),
    # match end-of-life, end-of-standard-support, out-of-standard-support anywhere in the line
    "version": re.compile(
        r"(^|.*\s+)(end[\s-]of[\s-]life|end[\s-]of[\s-]standard[\s-]support|out[\s-]of[\s-]standard[\s-]support)($|\s+|\,.*)",
    ),
}

# Only releases present in this mapping will be output by the driver, so it must be maintained with new releases.
ubuntu_version_names = {
    "precise": "12.04",
    # 'precise/esm': '12.04',
    "quantal": "12.10",
    "raring": "13.04",
    "trusty": "14.04",
    "utopic": "14.10",
    "vivid": "15.04",
    "wily": "15.10",
    "xenial": "16.04",
    "yakkety": "16.10",
    "zesty": "17.04",
    "artful": "17.10",
    "bionic": "18.04",
    "cosmic": "18.10",
    "disco": "19.04",
    "eoan": "19.10",
    "focal": "20.04",
    "groovy": "20.10",
    "hirsute": "21.04",
    "impish": "21.10",
    "jammy": "22.04",
    "kinetic": "22.10",
    "lunar": "23.04",
    "mantic": "23.10",
    "noble": "24.04",
    "oracular": "24.10",
    "plucky": "25.04",
}

# driver workspace
driver_workspace = None


@dataclass(frozen=True)
class DistroPkg:
    distro: str
    pkg: str


class JsonifierMixin:
    def json(self):
        jsonified = {}
        for k, v in vars(self).items():
            if k[0] != "_":
                if isinstance(v, (list, set)):
                    jsonified[k] = [x.json() if hasattr(x, "json") and callable(x.json) else x for x in v]
                elif isinstance(v, dict):
                    jsonified[k] = {x: y.json() if hasattr(y, "json") and callable(y.json) else y for x, y in v.items()}
                elif hasattr(v, "json"):
                    jsonified[k] = v.json()
                else:
                    jsonified[k] = v
        return jsonified


# Driver V2 Design Changes:
# Because the ubuntu data is removed when a release hits EOL (patch records changed from a state to 'ignored (reached end-of-life)'),
# the way to keep the actual state for all cves is to track the last update prior to EOL, which requires traversing revision history.
# The v2 driver does this to maintain an accurate state
# for each cve. It will be very slow on initial run but can be optimized by pre-populating the cache of merged state


class Vulnerability(JsonifierMixin):
    """
    Class representing the record to be returned. Uses strange capitalization
    to be backwards compatible in the json output with previous version of feed data.
    """

    def __init__(self):
        self.Name = None
        self.NamespaceName = None
        self.Description = ""
        self.Severity = None
        self.Metadata = None
        self.Link = None
        self.FixedIn = []


class FixedIn(JsonifierMixin):
    """
    Class representing a fix record for return back to the service from the driver. The semantics of the version are:
    "None" -> Package is vulnerable and no fix available yet
    ! "None" -> Version of package with a fix for a vulnerability. Assume all older versions of the package are vulnerable.

    """

    def __init__(self):
        self.Name = None
        self.NamespaceName = None
        self.VersionFormat = None
        self.Version = None
        self.VendorAdvisory = None


class Severity(enum.IntEnum):
    Unknown = 1
    Negligible = 2
    Low = 3
    Medium = 4
    High = 5
    Critical = 6

    def json(self):
        return self.name


@dataclass
class Patch:
    distro: str
    status: str
    version: str
    package: str | None = None
    priority: str | None = None


@dataclass
class CVEFile:
    # Keep this naming for now to preserve compatibility with legacy
    name: str
    priority: str = "Unknown"
    patches: list[Patch] = field(default_factory=list)
    ignored_patches: list[Patch] = field(default_factory=list)
    git_last_processed_rev: str | None = None
    references: list[str] | None = None
    description: str | None = None

    @staticmethod
    def from_dict(d: dict[str, Any]):
        name = d.get("name", d.get("Name", d.get("candidate", d.get("Candidate"))))
        priority = d.get("priority", d.get("Priority", "Unknown"))
        patches = [Patch(**p) for p in d.get("patches", [])]
        ignored_patches = [Patch(**p) for p in d.get("ignored_patches", [])]
        git_last_processed_rev = d.get("git_last_processed_rev")
        references = d.get("references", d.get("References"))
        description = d.get("description", d.get("Description"))
        return CVEFile(
            name=name,
            priority=priority,
            patches=patches,
            ignored_patches=ignored_patches,
            git_last_processed_rev=git_last_processed_rev,
            references=references,
            description=description,
        )


def check_header(expected: str, lines: list[str]):
    """
    Check that the first element of the lines list matches expected + ':'. And pop it off if found. Else raise exception.
    :param expected: The str value of the header prefix expected to be found on the first line
    :param lines: the list of lines, which will be modified on success by removing the first entry if it matches
    :return: None
    """

    # This is not a full match, only prefix to handle cases where the header isn't quite within spec (e.g. notes added etc).
    if not lines[0].strip().startswith(expected + ":"):
        raise ValueError(f"Expected header {expected}, found {lines[0]}")
    lines.pop(0)


def parse_list(header: str, lines: list[str]) -> list[str]:
    """
    Greedy parser that consumes lines if lines[0] is the header until the first empty line.

    :param header:
    :param lines:
    :return:
    """

    check_header(header, lines)

    refs: list[str] = []
    while lines:
        if _indent_line_regex.match(lines[0]):
            refs.append(lines.pop(0).strip())
        else:
            break

    return refs


def parse_patch(header: str, lines: list[str]) -> list[Patch]:  # noqa: C901
    """
    Parse a patch spec of the form:
    Patches_<packagename>:
    [<Priority_<packagename>: <priority>] (these are optional)
    [debdiff: URL]
    [vendor: URL]
    [upstream: URL]
    [patch: URL]
    <distro1>_<packagename>: <status> <status details>
    <distro2>_<packagename>: <status> <status details>
    ...
    <distroN>_<packagename>: <status> <status details>
    <newline>

    :param header:
    :param lines:
    :return:
    """
    if lines[0].startswith("Patches"):
        check_header(header, lines)

    patches: list[Patch] = []
    priority: str | None = None
    while lines:
        line = lines[0].strip()
        if not line:
            break

            # Check if this is a "Priority_<pkg>: <priority>" line
        pkg_priority = _package_priority_regex.match(line)
        if pkg_priority:
            pkg = pkg_priority.group(1)
            priority = pkg_priority.group(2)
            for patch in filter(lambda x: x.package == pkg, patches):
                patch.priority = priority
            lines.pop(0)
            continue
        else:  # noqa: RET507
            match = _patches_regex.match(line)
            if match:
                version = None
                state = match.group(3)
                status_match = _patch_state_regex.match(match.group(3))
                if status_match and status_match.group(1):
                    state = status_match.group(1)
                    if state in patch_states:
                        version = status_match.group(2)
                        if version:
                            version = version.strip()
                            if version[0] == "(" and version[-1] == ")":
                                version = version[1:-1]

                        p = Patch(distro=match.group(1), package=match.group(2), status=state, version=version, priority=priority)
                        patches.append(p)
                lines.pop(0)
            else:
                break

    return patches


def parse_simple_keyvalue(expected_key: str, lines: list[str]) -> str:
    """
    Parse a <Header>: <value> line and any skip ahead any empty lines following

    :param header: expected key
    :param lines: lines of input
    :return: tuple (value, line count consumed)
    """

    tokens = lines[0].split(":", 1)
    if len(tokens) != 2:
        raise ValueError(f"Could not parse {lines[0]} as key: value")

    key = tokens[0]
    value = tokens[1].strip()

    if key != expected_key:
        raise ValueError(f"Did not find expected key {expected_key}. Found {key}")

    lines.pop(0)

    return value


def parse_multiline_keyvalue(header: str, lines: list[str]) -> str:
    """
    Parse a header plus multiple lines (to an empty line) into a single string, stripping newlines

    :param header:
    :param lines:
    :return:
    """
    check_header(header, lines)

    content: list[str] = []
    while lines and _indent_line_regex.match(lines[0]):
        content.append(lines.pop(0).strip())

    return " ".join(content)


def get_patch_section(header_line: str) -> str | None:
    match = _patches_header_regex.match(header_line)
    if match:
        return match.group(1)
    return None


def check_release(releasename: str) -> bool:
    """
    Returns true if the releasename is one we care about.

    :param releasename: str name of the release
    :return: bool, True if release name is known, False if not
    """
    return releasename in ubuntu_version_names


def check_state(state: str) -> bool:
    """
    Returns true if the state is one we should process.
    :param state:l
    :return: bool, True if state is valid, False if not
    """
    return state and patch_states.get(state)


def check_patch(patch_record: Patch) -> bool:
    """
    Filtering test for a patch record to determine if this patch record indicates:
    * The release is affected
    * The release is one we are tracking

    :param patch_record: dict for a patch
    :return: True if record should be used, False if not
    """

    if not patch_record:
        return False

    # Distro filter step
    return check_release(patch_record.distro) and check_state(patch_record.status)


def check_merge(patch_record: Patch) -> bool:
    """
    Test a patch record to see if it meets the merge criteria

    :param patch_record:
    :return:
    """

    if not patch_record:
        return False

    patch_dict = asdict(patch_record)

    # perform regex match only if the value is non-null and non-empty
    return all(patch_dict.get(key, None) and re.match(regex, patch_dict.get(key)) for key, regex in patch_merge_criteria.items())


def parse_cve_file(cve_id: str, content_lines: list[str]) -> CVEFile:
    """
    Parses a single CVE file into a dict representation suitable for further processing.

    :param lines:
    :return: dict version of the cve record
    """

    parsed = CVEFile(name=cve_id)

    # Copy to avoid modifying the passed param directly since the parsing is greedy.
    lines = copy.deepcopy(content_lines)
    while lines:
        line = lines[0].strip()
        if not line or line.startswith("#"):
            # Skip empties and comments
            lines.pop(0)
        else:
            section = line.split(":", 1)[0]
            if section == "Candidate":
                parsed.name = parse_simple_keyvalue(section, lines)
            elif section == "References":
                parsed.references = parse_list(section, lines)
            elif section == "Description":
                parsed.description = parse_multiline_keyvalue(section, lines)
            elif section == "Priority":
                parsed.priority = parse_simple_keyvalue(section, lines)
            else:
                patch_name = get_patch_section(section)
                p_match = _patches_regex.match(line)
                if patch_name:  # noqa: SIM114
                    patches = parse_patch(section, lines)
                    parsed.patches += patches
                    continue
                elif p_match and map_namespace(p_match.group(1)):  # noqa: RET507
                    patches = parse_patch(section, lines)
                    parsed.patches += patches
                    continue
                else:
                    # Consume the line. No match to parse.
                    lines.pop(0)

    return parsed


def map_namespace(release_name: str) -> str | None:
    """
    Returns a namespace name (ubuntu:<version>) where version is a numeric id instead of release name
    e.g. map_namespace('vivid') -> ubuntu:15.04
    :param release_name:
    :return:
    """

    dist = ubuntu_version_names.get(release_name)
    if dist:
        return ":".join([namespace, dist])

    return None


def parse_severity_from_priority(cve: CVEFile) -> Severity:
    severity = cve.priority.capitalize()
    if severity in {"Untriaged"}:
        return Severity.Unknown
    return getattr(Severity, severity)


def map_parsed(parsed_cve: CVEFile, logger: logging.Logger | None = None):  # noqa: C901, PLR0912
    """
    Maps a parsed CVE dict into a Vulnerability object.

    Each parsed CVE dict contains data for multiple distro releases and packages.
    The returned output is a set of Vulnerability objects.

    The overall approach is to emit a CVE record for each namespace that has an entry for a patch, even if that entry is a Does Not Effect.
    The output can be Vulnerability objects with no FixedIn records.

    :param parsed_cve:
    :return: set of Vulnerability objects
    """

    if not logger:
        logger = logging.getLogger(__name__)

    # Map keyed by namespace name
    vulns = {}
    if not (parsed_cve.name):
        logger.error(f"could not find a Name for parsed cve: {asdict(parsed_cve)}")
        return []

    for p in parsed_cve.patches:
        namespace_name = map_namespace(p.distro)

        # Build the CVE record even if no fixedIn record
        r = vulns.get(namespace_name)
        if not r:
            if not namespace_name:
                continue

            r = Vulnerability()

            try:
                r.Severity = parse_severity_from_priority(parsed_cve)
            except AttributeError:
                logger.warning(
                    f"setting unknown severity on {parsed_cve.name} due to unsupported priority value {parsed_cve.priority}",
                )
                r.Severity = Severity.Unknown
            except Exception:
                logger.exception(f"setting unknown severity on {parsed_cve.name} due to exception parsing severity from priority")
                r.Severity = Severity.Unknown

            r.Name = parsed_cve.name
            r.Metadata = {}
            r.Link = ubuntu_cve_url.format(r.Name)
            r.FixedIn = []
            r.NamespaceName = namespace_name
            vulns[namespace_name] = r

        # If the patch status is one we care about, make the FixedIn record, else skip it but create CVE records
        # We currently want to mark end-of-support records with no previously known fix as vulnerable, hence the
        # or check_merge step here.
        if check_state(p.status) or check_merge(p):
            pkg = FixedIn()
            pkg.Name = p.package

            # If there is a version indicating a fix use it, else 'None' is special keyword for no-fix-available
            if p.status == "released":
                # Can do version format check here, but requires code from anchore-engine
                # anchore_engine.services.policy_engine.engine.util.deb.DpkgVersion.from_string(p.get('status'))
                pkg.Version = p.version
                if pkg.Version is None:
                    logger.debug(
                        'found CVE {} in ubuntu version {} with "released" status for pkg {} but no version for release. Released patches should have version info, but missing in source data. Marking package as not vulnerable'.format(  # noqa: E501, G001, UP032
                            r.Name,
                            r.NamespaceName,
                            pkg.Name,
                        ),
                    )
                    continue
                    # Strange condition where a release was done but no version found. In this case, we'll omit the FixedIn record.

            else:
                pkg.Version = "None"
                # Set NoAdvisory to true so that `wont-fix` status gets set on
                # out of support entries
                if p.status == "ignored":
                    pkg.VendorAdvisory = {"NoAdvisory": True}

            if not pkg.VendorAdvisory:
                pkg.VendorAdvisory = {"NoAdvisory": False}

            pkg.VersionFormat = "dpkg"
            pkg.NamespaceName = namespace_name
            r.FixedIn.append(pkg)

            # Check for max priority of all packages with it set
            if p.priority:
                pkg_sev = getattr(Severity, p.priority.capitalize())
                r.Severity = max(pkg_sev, r.Severity)

    return set(vulns.values())


def filter_resolved_patches(cve: CVEFile, dpt_list: list[DistroPkg]) -> dict[DistroPkg, Patch]:
    """
    Filter patch records from the cve dictionary that match the package and distribution of the items in the list but don't satisfy merge criteria

    :param cve_dict: dict version of a cve record, like the return value of parse_cve_file()
    :param dpt_list: list of DistroPkg objects
    :return:
    """
    filtered_map = {}
    for dpt in dpt_list:
        matched_p = next(
            (p for p in cve.patches if dpt.distro == p.distro and dpt.pkg == p.package and not check_merge(p)),
            None,
        )
        if matched_p:
            filtered_map[dpt] = matched_p

    return filtered_map


def filter_merged_patches(cve: CVEFile, dpt_list: list[DistroPkg]) -> dict[DistroPkg, Patch]:
    """
    Filter patch records from the cve dictionary that match the package and distribution of the items in the list

    :param cve_dict: dict version of a cve record, like the return value of parse_cve_file()
    :param dpt_list: list of DistroPkg objects
    :return:
    """
    filtered_map: dict[DistroPkg, Patch] = {}
    for dpt in dpt_list:
        matched_p = next(
            (p for p in cve.patches if dpt.distro == p.distro and dpt.pkg == p.package),
            None,
        )
        if matched_p:
            filtered_map[dpt] = matched_p

    return filtered_map


class Parser:
    __payload__ = Vulnerability

    _bzr_src = "https://launchpad.net/ubuntu-cve-tracker"
    _git_src_url = "git://git.launchpad.net/ubuntu-cve-tracker"
    _bzr_to_git_transition_commit = "dc3f64a0dfe6b1780240ff115d8a0a1b23fd00b4"

    _active_cve_dir = "active"
    _retired_cve_dir = "retired"
    _ignored_cve_dir = "ignored"  # May need to use this later if CVEs transition from one to another
    _vc_working_dir = "ubuntu-cve-tracker"
    _normalized_cve_dir = "normalized-cve-data"
    _last_processed_rev_file = "bzr-last-processed-rev"
    _last_processed_rev_file_git = "git-last-processed-rev"

    data_filename_regex = re.compile(r"(active|retired|ignored)/CVE-[0-9]{4}-[0-9]+")

    # Revision 9000 on the sec bzr repo before trusty eol. revno 9000 on 2015-01-27.

    def __init__(  # noqa: PLR0913
        self,
        workspace: Workspace,
        logger: logging.Logger | None = None,
        additional_versions: dict[str, str] | None = None,
        enable_rev_history: bool = True,
        max_workers: int = default_max_workers,
        git_url: str = default_git_url,
        git_branch: str = default_git_branch,
    ):
        self.vc_workspace = os.path.join(workspace.input_path, self._vc_working_dir)
        # TODO: tech debt: this should use the results workspace with the correct schema-aware envelope
        self.norm_workspace = os.path.join(workspace.input_path, self._normalized_cve_dir)
        if not logger:
            logger = logging.getLogger(self.__class__.__name__)
        self.logger = logger
        self.git_url = git_url
        self.git_branch = git_branch
        self.urls = [self.git_url]
        self.git_wrapper = GitWrapper(source=self.git_url, branch=self.git_branch, checkout_dest=self.vc_workspace, logger=logger)

        if additional_versions:
            ubuntu_version_names.update(additional_versions)
        self.enable_rev_history = enable_rev_history
        self._max_workers = max_workers

    def fetch(self, skip_if_exists=False):
        # setup merged workspace
        if not os.path.exists(self.norm_workspace):
            os.makedirs(self.norm_workspace)

        # sync git repo
        self.git_wrapper.init_repo()

        # get the last processed revision if available and the current latest revision
        last_rev = self._load_last_processed_rev()
        current_rev = self.git_wrapper.get_current_rev()

        # process all the updates between last processed and current revisions
        self._process_data(self.vc_workspace, from_rev=last_rev, to_rev=current_rev)
        # save last processed revision to disk
        self._save_last_processed_rev(current_rev)

        # load merged state and map it to vulnerabilities
        self.logger.info("begin loading processed CVE content and transforming into vulnerabilities")
        for merged_cve in self._merged_cve_iterator():
            yield from map_parsed(merged_cve, self.logger)
        self.logger.info("finish loading processed CVE content and transforming into vulnerabilities")

    def _process_data(self, vc_dir: str, to_rev: str, from_rev: str | None = None):  # noqa: C901
        self.logger.info(f"processing data from git repository: {vc_dir}, from revision: {from_rev}, to revision: {to_rev}")

        self.git_wrapper.prepare_cve_revision_history()

        # gather a list of changed files if the last repo revision processed is available
        updated_paths = []
        deleted_ids = []
        if from_rev and to_rev and from_rev != to_rev:
            self.logger.debug(f"fetching changes to CVEs between revisions {from_rev} and {to_rev}")
            modified, removed = self.git_wrapper.get_merged_change_set(from_rev=from_rev, to_rev=to_rev)
            updated_paths = list(modified.values()) if modified else []
            deleted_ids = list(removed.keys()) if removed else []
            self.logger.info(f"detected {len(updated_paths)} CVE updates (add/modify/rename)")
            self.logger.info(f"detected {len(deleted_ids)} CVE deletions")

        # Load cves from active and retired directories and spool merged state to disk
        # note: this is an IO bound operation, so a thread pool will suffice for now
        # but look to a process pool if this becomes a bottleneck
        proc_exception = None
        self.logger.info("begin processing updates")
        with concurrent.futures.ThreadPoolExecutor(max_workers=self._max_workers) as executor:

            def worker(fn, cve_id: str, *args, **kwargs):
                try:
                    return fn(cve_id, *args, **kwargs)
                except:
                    self.logger.exception(f"error processing {cve_id}")
                    raise

            futures = []

            for d in [self._active_cve_dir, self._retired_cve_dir]:
                cve_dir = os.path.join(vc_dir, d)
                for cve_id in sorted(filter(lambda x: _cve_filename_regex.match(x), os.listdir(cve_dir))):
                    f = os.path.join(cve_dir, cve_id)
                    cve_rel_path = "/".join([d, cve_id])

                    future = executor.submit(worker, self._process_cve, cve_id, cve_rel_path, f, to_rev, updated_paths)
                    futures.append(future)

            # wait for all the futures to complete
            done, not_done = concurrent.futures.wait(futures, return_when=concurrent.futures.FIRST_EXCEPTION)

            if len(done) > 0 and len(done) != len(futures):
                future = done.pop()
                proc_exception = future.exception()
                if proc_exception:
                    self.logger.error(f"one task failed with: {proc_exception} (shutting down)")

                    # cancel any scheduled tasks
                    for future in futures:
                        future.cancel()

        if proc_exception:
            raise proc_exception

        self.logger.info("finish processing updates")

        # Remove merged state of deleted cves
        self.logger.info("begin processing deletes")
        for cve_id in deleted_ids:
            self.logger.debug(f"{cve_id} is no longer relevant, deleting merged CVE state if any")
            self._delete_merged_cve(cve_id)
        self.logger.info("finish processing deletes")

    def _process_cve(
        self,
        cve_id: str,
        cve_rel_path: str,
        f: str,
        to_rev: str,
        updated_paths: list[str],
    ) -> CVEFile | None:
        self.logger.debug(f"begin processing {cve_id} to rev {to_rev}")

        if cve_rel_path in updated_paths:
            # merge cves updated since last revision or all if the last processed revision is not available
            # self.logger.debug("CVE updated since last run, processing {}".format(cve_rel_path))
            result = self._merge_cve(cve_id, cve_rel_path, f, to_rev)
        elif not self._merged_cve_exists(cve_id):
            # merge may be required since the saved state is not found
            # self.logger.debug("CVE merged state not found, processing {}".format(cve_rel_path))
            result = self._merge_cve(cve_id, cve_rel_path, f, to_rev)
        else:
            # merge may be required if new distros were added
            # self.logger.debug("reprocessing merged CVE {}".format(cve_rel_path))
            result = self._reprocess_merged_cve(cve_id, cve_rel_path)

        self.logger.debug(f"finish processing {cve_id} to rev {to_rev}")
        return result

    def _load_last_processed_rev(self):
        last_processed_rev_path = os.path.join(self.norm_workspace, self._last_processed_rev_file_git)

        if os.path.exists(last_processed_rev_path):
            with open(last_processed_rev_path) as fp:
                return fp.readline().strip()
        else:
            last_processed_rev_path = os.path.join(self.norm_workspace, self._last_processed_rev_file)
            if os.path.exists(last_processed_rev_path):
                return self._bzr_to_git_transition_commit
            return None

    def _save_last_processed_rev(self, revno: str):
        last_processed_rev_path = os.path.join(self.norm_workspace, self._last_processed_rev_file_git)

        with open(last_processed_rev_path, "w") as f:
            f.write(f"{revno}")

    def _load_merged_cve(self, cve_id: str) -> CVEFile | None:
        if os.path.exists(os.path.join(self.norm_workspace, cve_id)):
            with open(os.path.join(self.norm_workspace, cve_id)) as fp:
                cve_json = orjson.loads(fp.read())
                return CVEFile.from_dict(cve_json)

        return None

    def _save_merged_cve(self, cve_id: str, merged_cve: CVEFile):
        filepath = os.path.join(self.norm_workspace, cve_id)
        with open(filepath, "wb") as f:
            self.logger.trace(f"writing record to {filepath!r}")  # type: ignore  # noqa: PGH003
            f.write(orjson.dumps(asdict(merged_cve), f))  # type: ignore  # noqa: PGH003

    def _delete_merged_cve(self, cve_id):
        if os.path.exists(os.path.join(self.norm_workspace, cve_id)):
            os.remove(os.path.join(self.norm_workspace, cve_id))

    def _merged_cve_iterator(self) -> Generator[CVEFile, None, None]:
        for cve_id in filter(lambda x: _cve_filename_regex.match(x), os.listdir(self.norm_workspace)):
            with open(os.path.join(self.norm_workspace, cve_id)) as fp:
                cve = orjson.loads(fp.read())
                yield CVEFile.from_dict(cve)

    def _merged_cve_exists(self, cve_id):
        return os.path.exists(os.path.join(self.norm_workspace, cve_id))

    def _reprocess_merged_cve(self, cve_id: str, cve_rel_path: str):
        """
        Assumes that a normalized state exists for cve and processes only ignored patches, mainly for handling new distros

        :param cve_id:
        :param cve_rel_path:
        :return:
        """
        self.logger.debug(f"reprocessing merged CVE {cve_rel_path}")
        saved_state = self._load_merged_cve(cve_id)

        if not saved_state:
            self.logger.debug(f"no saved state found for {cve_id}")
            return None

        # reprocess only ignored patches
        merged_patches, ignored_patches, to_be_merged_map = self._categorize_patches(saved_state.ignored_patches)

        # Found patches that can be merged and or can't be resolved from the saved state, could be a new namespace
        if merged_patches or to_be_merged_map:
            self.logger.debug("found unresolved patches in previously merged state, could be a new distro")
            # Process revision history for eol-ed packages that need to be merged
            if to_be_merged_map:
                if self.enable_rev_history:
                    self.logger.debug(f"attempting to resolve patches using revision history for {cve_rel_path}")
                    (
                        resolved_patches,
                        pending_dpt_list,
                        cve_latest_rev,
                    ) = self._resolve_patches_using_history(
                        cve_id=cve_id,
                        cve_rel_path=cve_rel_path,
                        to_be_merged_dpt_list=list(to_be_merged_map.keys()),
                        priority=saved_state.priority,
                    )
                    merged_patches.extend(resolved_patches)
                    if pending_dpt_list:
                        self.logger.debug(
                            "exhausted all revisions for {} but could not resolve patches: {}".format(  # noqa: G001, UP032
                                cve_rel_path,
                                [to_be_merged_map[x] for x in pending_dpt_list],
                            ),
                        )
                        merged_patches.extend([to_be_merged_map[x] for x in pending_dpt_list])

                    del resolved_patches[:]
                    del pending_dpt_list[:]

                    if cve_latest_rev:
                        saved_state.git_last_processed_rev = cve_latest_rev
                else:
                    self.logger.debug("revision history processing is disabled. Merging unresolved patches as they are")
                    merged_patches.extend(to_be_merged_map.values())

            # pulling this outside of the revision history block for fixing ENTERPRISE-195. saved state should be updated if there are mergeable or to-be-merged packages  # noqa: E501
            # there might already be resolved patches, extend it with merged patches, don't overwrite it
            if saved_state.patches:
                saved_state.patches.extend(merged_patches)
            else:
                saved_state.patches = merged_patches

            # overwrite ignored patches since its the final list
            saved_state.ignored_patches = ignored_patches

            # save the merged cve state to disk before returning
            self._save_merged_cve(cve_id, saved_state)
        else:  # No patches that require resolution
            pass

        return saved_state

    @staticmethod
    def _categorize_patches(patch_list: list[Patch]) -> tuple[list[Patch], list[Patch], dict[DistroPkg, Patch]]:
        """
        Takes in a list of patches and splits them into merged and ignored lists and a dictionary of patches that need further resolution
        :param patch_list:
        :return:
        """

        to_be_merged_map: dict[DistroPkg, Patch] = {}
        merged_patches: list[Patch] = []
        ignored_patches: list[Patch] = []

        for p in patch_list:
            namespace_name = map_namespace(p.distro)

            if not namespace_name:
                ignored_patches.append(p)
            elif namespace_name and check_merge(p):
                to_be_merged_map[DistroPkg(distro=p.distro, pkg=p.package)] = p
            else:
                # Patch does not meet criteria for resolution - either its not eol-ed or in a namespace we care about
                merged_patches.append(p)

        return merged_patches, ignored_patches, to_be_merged_map

    def _merge_cve(self, cve_id: str, cve_rel_path: str, cve_abs_path: str, repo_current_rev: str):
        """
        Parses the contents of an ubuntu security tracker CVE file and returns the normalized data as a dictionary.
        If the CVE has any patches/fixes that meet the merge criteria, the revision history of the CVE is traversed
        in the most recent to least recent order to compute the last known state of the patch.

        :param cve_abs_path:
        :param repo_current_rev:
        :return:
        """
        self.logger.debug(f"merging CVE {cve_rel_path}")

        with open(cve_abs_path) as cve_file:
            raw_content = cve_file.readlines()
            parsed_cve = parse_cve_file(cve_id, raw_content)

        cve_latest_rev = None
        merged_patches, ignored_patches, to_be_merged_map = self._categorize_patches(parsed_cve.patches)

        # Found patches that can't be resolved from the input data, could be the first time the driver is running
        if to_be_merged_map:
            # lookup saved state
            saved_cve = self._load_merged_cve(cve_id)

            if self.enable_rev_history:
                (
                    resolved_patches,
                    pending_dpt_list,
                    cve_latest_rev,
                ) = self._resolve_patches_using_history(
                    cve_id=cve_id,
                    cve_rel_path=cve_rel_path,
                    to_be_merged_dpt_list=list(to_be_merged_map.keys()),
                    priority=parsed_cve.__dict__.get("Priority", "Unknown"),
                    saved_state=saved_cve,
                )
                merged_patches.extend(resolved_patches)
                if pending_dpt_list:
                    self.logger.debug(
                        f"exhausted all revisions for {cve_rel_path} but could not resolve patches: {[to_be_merged_map[x] for x in pending_dpt_list]}",  # noqa: E501
                    )
                    merged_patches.extend([to_be_merged_map[x] for x in pending_dpt_list])

                del resolved_patches[:]
                del pending_dpt_list[:]

            else:  # merge with saved state if any  # noqa: PLR5501
                if saved_cve:
                    self.logger.debug("revision history processing is disabled. Resolving patches using saved cve state")
                    rev_matched_map = filter_merged_patches(saved_cve, list(to_be_merged_map.keys()))
                    # merge resolved and unresolved patches
                    merged_patches.extend(list(rev_matched_map.values()))
                    merged_patches.extend([to_be_merged_map[x] for x in to_be_merged_map if x not in rev_matched_map])

                    rev_matched_map.clear()
                else:
                    self.logger.debug(
                        "revision history processing is disabled and no saved state found. Skipping patch resolution",
                    )
                    merged_patches.extend(list(to_be_merged_map.values()))

        else:  # No patches that require resolution
            pass

        parsed_cve.patches = merged_patches
        parsed_cve.ignored_patches = ignored_patches
        parsed_cve.git_last_processed_rev = cve_latest_rev if cve_latest_rev else repo_current_rev

        # save the merged cve state to disk before returning
        self._save_merged_cve(cve_id, parsed_cve)

        return parsed_cve

    def _resolve_patches_using_history(  # noqa: C901, PLR0912, PLR0915
        self,
        cve_id: str,
        cve_rel_path: str,
        to_be_merged_dpt_list: list[DistroPkg],
        priority: str | None = None,
        saved_state: CVEFile | None = None,
    ):
        t = time.time()
        self.logger.debug(f"processing CVE revision history for: {cve_rel_path}")

        # setup metrics
        metrics = {
            "revs_processed": 0,
            "previous_state_counter": {},
            "time_elapsed": 0,
            "severity": priority,
        }

        # initialize
        resolved_patches: list[Patch] = []

        # copy the keys so the original ds is not modified
        pending_dpt_list: list[DistroPkg] = copy.deepcopy(to_be_merged_dpt_list)

        # last processed commit
        saved_cve_last_processed_rev = saved_state.git_last_processed_rev if saved_state and saved_state.git_last_processed_rev else None

        # fetch log of revision history for this file, its in the most recent - least recent order
        since_revs = self.git_wrapper.get_revision_history(cve_id, cve_rel_path, saved_cve_last_processed_rev)

        # make note of the first revision, its the latest revision and one we already processed
        cve_latest_rev = since_revs[0].sha if since_revs and since_revs[0] else None

        # first process all revisions before last processed revision
        if since_revs and len(since_revs) > 1:
            # process only previous revisions, not current
            for rev in since_revs[1:]:
                # getting revision content is expensive, so check things to merge first
                if not pending_dpt_list:
                    break  # No more merging required, stop processing revs!

                # process revision: fetch content, parse, filter patches that need to be merged and so on
                rev_raw_content = self.git_wrapper.get_content(rev)

                # parse the cve and filter based on merge criteria
                rev_cve = parse_cve_file(cve_id, rev_raw_content)
                rev_matched_map = filter_resolved_patches(rev_cve, pending_dpt_list)
                resolved_patches.extend(list(rev_matched_map.values()))
                pending_dpt_list = [x for x in pending_dpt_list if x not in rev_matched_map]

                # metrics for processed revision
                metrics["revs_processed"] += 1
                for rev_p in rev_matched_map.values():
                    rev_p_ns = map_namespace(rev_p.distro)
                    patch_status = rev_p.status if not None else "Unknown"
                    patch_distro = rev_p.distro if not None else "Unknown"
                    if patch_status not in metrics["previous_state_counter"]:
                        metrics["previous_state_counter"][patch_status] = {rev_p_ns: 0}
                    if patch_distro not in metrics["previous_state_counter"][patch_status]:
                        metrics["previous_state_counter"][patch_status][rev_p_ns] = 0

                    metrics["previous_state_counter"][patch_status][rev_p_ns] += 1

                # free up after processing
                rev_matched_map.clear()
                del rev_cve
                del rev_raw_content[:]
        else:
            # no revs for processing
            self.logger.debug("no previous revisions found")

        # then merge with saved state if there are things that still need to be merged. This is a one time thing to short circuit fetching revs
        if pending_dpt_list and saved_state:
            self.logger.debug("resolving patches using saved cve state before processing any more revisions")
            rev_matched_map = filter_merged_patches(saved_state, pending_dpt_list)
            resolved_patches.extend(list(rev_matched_map.values()))
            pending_dpt_list = [x for x in pending_dpt_list if x not in rev_matched_map]

            # free up after processing
            rev_matched_map.clear()

            # Last, merging with saved state may or may not resolve patch state.
            # If a new namespace is introduced, its possible the saved cve is not fully resolved.
            # So continue with process revision: fetch content, parse, filter patches that need to be merged and so on
            if pending_dpt_list:
                # fetch the entire revision history for this file, needed for computing the merge
                self.logger.debug("unresolved patches found even after merge with saved state, walking entire revision history")
                all_revs = self.git_wrapper.get_revision_history(cve_id, cve_rel_path, None)
                before_revs = all_revs[len(since_revs) :] if since_revs else all_revs

                for rev in before_revs:
                    # getting revision content is expensive, so check things to merge first
                    if not pending_dpt_list:
                        break  # No more merging required, stop processing revs!

                    # process revision: fetch content, parse, filter patches that need to be merged and so on
                    rev_raw_content = self.git_wrapper.get_content(rev)

                    # parse the cve and filter based on merge criteria
                    rev_cve = parse_cve_file(cve_id, rev_raw_content)
                    rev_matched_map = filter_resolved_patches(rev_cve, pending_dpt_list)
                    resolved_patches.extend(list(rev_matched_map.values()))
                    pending_dpt_list = [x for x in pending_dpt_list if x not in rev_matched_map]

                    # metrics for processed revision
                    metrics["revs_processed"] += 1
                    for rev_p in rev_matched_map.values():
                        rev_p_ns = map_namespace(rev_p.distro)
                        if rev_p.status not in metrics["previous_state_counter"]:
                            metrics["previous_state_counter"][rev_p.status] = {rev_p_ns: 0}
                        if rev_p.distro not in metrics["previous_state_counter"][rev_p.status]:
                            metrics["previous_state_counter"][rev_p.status][rev_p_ns] = 0

                        metrics["previous_state_counter"][rev_p.status][rev_p_ns] += 1

                    # free up after processing
                    rev_matched_map.clear()
                    del rev_cve
                    del rev_raw_content[:]
            else:
                self.logger.debug("Merge with saved state resolved all relevant patches")

        metrics["time_elapsed"] = int((time.time() - t) * 1000) / float(1000)
        self.logger.trace(f"metrics from processing revision history for {cve_rel_path}: {metrics}")

        return resolved_patches, pending_dpt_list, cve_latest_rev

    def get(self, skip_if_exists=False):
        for i in self.fetch(skip_if_exists=skip_if_exists):
            yield i.NamespaceName, i.Name, i.json()



from __future__ import annotations

import shutil
from requests.exceptions import HTTPError
from unittest.mock import Mock

import pytest
from vunnel import result, workspace
from vunnel.utils.http_wrapper import requests
from vunnel.providers.amazon import Config, Provider, parser


class TestParser:
    def test_rss_parsing(self, tmpdir, helpers):
        mock_data_path = helpers.local_dir("test-fixtures/mock_rss")
        p = parser.Parser(workspace=workspace.Workspace(tmpdir, "test", create=True))
        summaries = p._parse_rss(mock_data_path)

        assert isinstance(summaries, list)
        assert len(summaries) > 0  # TODO: what is the right answer?

        for alas in summaries:
            assert isinstance(alas, parser.AlasSummary)
            assert alas.id is not None
            assert alas.url is not None
            assert alas.sev.lower() in parser.severity_map

        # TODO: beef up these assertions (should cover the full data shape)

    def test_html_parsing(self, helpers):
        new_packages = [
            "java-1.8.0-openjdk-javadoc-1.8.0.161-0.b14.amzn2.noarch",
            "java-1.8.0-openjdk-javadoc-zip-1.8.0.161-0.b14.amzn2.noarch",
            "java-1.8.0-openjdk-javadoc-debug-1.8.0.161-0.b14.amzn2.noarch",
            "java-1.8.0-openjdk-javadoc-zip-debug-1.8.0.161-0.b14.amzn2.noarch",
            "java-1.8.0-openjdk-1.8.0.161-0.b14.amzn2.src",
            "java-1.8.0-openjdk-1.8.0.161-0.b14.amzn2.x86_64",
            "java-1.8.0-openjdk-debug-1.8.0.161-0.b14.amzn2.x86_64",
            "java-1.8.0-openjdk-headless-1.8.0.161-0.b14.amzn2.x86_64",
            "java-1.8.0-openjdk-headless-debug-1.8.0.161-0.b14.amzn2.x86_64",
            "java-1.8.0-openjdk-devel-1.8.0.161-0.b14.amzn2.x86_64",
            "java-1.8.0-openjdk-devel-debug-1.8.0.161-0.b14.amzn2.x86_64",
            "java-1.8.0-openjdk-demo-1.8.0.161-0.b14.amzn2.x86_64",
            "java-1.8.0-openjdk-demo-debug-1.8.0.161-0.b14.amzn2.x86_64",
            "java-1.8.0-openjdk-src-1.8.0.161-0.b14.amzn2.x86_64",
            "java-1.8.0-openjdk-src-debug-1.8.0.161-0.b14.amzn2.x86_64",
            "java-1.8.0-openjdk-accessibility-1.8.0.161-0.b14.amzn2.x86_64",
            "java-1.8.0-openjdk-accessibility-debug-1.8.0.161-0.b14.amzn2.x86_64",
            "java-1.8.0-openjdk-debuginfo-1.8.0.161-0.b14.amzn2.x86_64",
        ]

        mock_data_path = helpers.local_dir("test-fixtures/mock_html")
        with open(mock_data_path) as fp:
            html_content = fp.read()

        p = parser.PackagesHTMLParser()
        p.feed(html_content)

        assert p.fixes is not None
        assert new_packages == p.fixes
        assert p.issue_overview_text is not None
        # TODO: beef up these assertions (should cover the full data shape)

    def test_get_pkg_name_version(self):
        f = parser.Parser.get_package_name_version("java-1.8.0-openjdk-javadoc-1.8.0.161-0.b14.amzn2.noarch")
        assert f.pkg == "java-1.8.0-openjdk-javadoc"
        assert f.ver == "1.8.0.161-0.b14.amzn2"

        f = parser.Parser.get_package_name_version("java-1.8.0-openjdk-1.8.0.161-0.b14.amzn2.src")
        assert f.pkg == "java-1.8.0-openjdk"
        assert f.ver == "1.8.0.161-0.b14.amzn2"

        f = parser.Parser.get_package_name_version("java-1.8.0-openjdk-debug-1.8.0.161-0.b14.amzn2.x86_64")
        assert f.pkg == "java-1.8.0-openjdk-debug"
        assert f.ver == "1.8.0.161-0.b14.amzn2"

        a = parser.Parser.get_package_name_version("java-1.8.0-openjdk-1.8.0.161-0.b14.amzn2.src")
        b = parser.Parser.get_package_name_version("java-1.8.0-openjdk-1.8.0.161-0.b14.amzn2.x86_64")
        assert a == b

    def test_get_alas_html_403(self, helpers, monkeypatch, tmpdir):
        # write a mock such that any http.get call will return a response with status code 403
        def mock_get(*args, **kwargs):
            return Mock(status_code=403)

        monkeypatch.setattr(requests, "get", mock_get)

        alas_file = tmpdir.join("alas.html")

        p = parser.Parser(workspace=workspace.Workspace(helpers.local_dir("test-fixtures"), "test", create=True))
        alas = p._get_alas_html("https://example.com", alas_file)
        assert alas is None

    def test_get_alas_html_raises_over_threshold(self, helpers, monkeypatch, tmpdir):
        # write a mock such that any http.get call will return a response with status code 403
        url = "https://example.com"

        def mock_get(*args, **kwargs):
            return Mock(status_code=403, url=url)

        monkeypatch.setattr(requests, "get", mock_get)

        alas_file = tmpdir.join("alas.html")

        p = parser.Parser(workspace=workspace.Workspace(helpers.local_dir("test-fixtures"), "test", create=True))
        p.max_allowed_alas_http_403 = 2

        # assert does not raise when at the threshold
        p.alas_403s = ["something"]
        p._get_alas_html(url, alas_file)
        assert p.alas_403s == ["something", url]

        # assert raises when above the threshold
        with pytest.raises(ValueError):
            p._get_alas_html(url, alas_file)

        assert p.alas_403s == ["something", url, url]


def test_provider_schema(helpers, disable_get_requests, monkeypatch):
    workspace = helpers.provider_workspace_helper(name=Provider.name())

    c = Config()
    c.runtime.result_store = result.StoreStrategy.FLAT_FILE
    p = Provider(root=workspace.root, config=c)

    mock_data_path = helpers.local_dir("test-fixtures/input")
    shutil.copytree(mock_data_path, workspace.input_dir, dirs_exist_ok=True)

    def mock_download(self, *args, **kwargs):
        pass

    monkeypatch.setattr(p.parser, "_download_rss", mock_download)

    p.update(None)

    assert 3 == workspace.num_result_entries()
    assert workspace.result_schemas_valid(require_entries=True)


def test_provider_via_snapshot(helpers, disable_get_requests, monkeypatch):
    workspace = helpers.provider_workspace_helper(
        name=Provider.name(),
        input_fixture="test-fixtures/input",
    )

    c = Config()
    # keep all of the default values for the result store, but override the strategy
    c.runtime.result_store = result.StoreStrategy.FLAT_FILE
    p = Provider(
        root=workspace.root,
        config=c,
    )

    def mock_download(_url, _file):
        return None

    monkeypatch.setattr(p.parser, "_download_rss", mock_download)

    p.update(None)

    workspace.assert_result_snapshots()
